import datetime as dt
import socket
import os
import cv2

import getpass
import urllib.request
import pyautogui
import base64

# ---------- files:
FILE_USERPASS = "~/.config/flashcam/.flashcam_upw"  # will extend
FILE_REDCROSS0 = "~/.config/flashcam/crossred"  # will extend
FILE_REDCROSS = "~/.config/flashcam/crossred"  # will extend



def get_passfile( videodev ):
    """
    only initial part of passfile.... useful for redcross
    """
    if "passfile" in locals():
        if passfile is None:
            print(f"i... nO passfile...trying {videodev} , {passfile}")
            passfile = videodev.strip("http://")
            print("i... TRYING", passfile)
    else:
        passfile = videodev.strip("http://")
        passfile = passfile.strip("/video")
        passfile = passfile.split(":")[0]

        ### WITH A HACK  -  I REDEFINE REDCROSS too
        ###FILE_REDCROSS = f"{FILE_REDCROSS0}_{passfile}.txt"

        # maybe not needed here
        #FILE_REDCROSS = get_FILE_REDCROSS(passfile)

    return passfile




def get_FILE_REDCROSS( videodev ):
    """
    used to save coordinates few times...
    """
    passfile = get_passfile( videodev)
    FILE_REDCROSS = f"{FILE_REDCROSS0}_{passfile}.txt"
    return FILE_REDCROSS




def get_stream( videodev="" ):# , FILE_USERPASS, FILE_REDCROSS,FILE_REDCROSS0 ):
    # global CTRL, SHIFT
    # localuser
    #global FILE_USERPASS, FILE_REDCROSS,FILE_REDCROSS0
    stream = None  # i return
    u, p = getpass.getuser(), "a"

    # this is bug.... never find passfile
    passfile = get_passfile( videodev )
    passfile = f"{FILE_USERPASS}_{passfile}"
    print(f"i... TRYING {videodev} PASS: {passfile}")

    nopass = True
    try:
        with open(os.path.expanduser(passfile)) as f:
            print("YES---> PASSWORD FILE  ", passfile)
            w1 = f.readlines()
            u = w1[0].strip()
            p = w1[1].strip()
            nopass = False
    except:
        print("NO PASSWORD FILE (gs) ")
        nopass = True
    if nopass:
        u = pyautogui.prompt("User:")
        p = pyautogui.prompt("User:")

        with open(os.path.expanduser(passfile), "w") as f:
            f.write(u)
            f.write("\n")
            f.write(p)
            f.write("\n")


    print("D... capturing from: /{}/".format(videodev))
    # cam = cv2.VideoCapture( videodev )
    # stream = urllib.request.urlopen( videodev )

    request = urllib.request.Request(videodev)
    print("D... USER/PASS", u, p)
    base64string = base64.b64encode(bytes("%s:%s" % (u, p), "ascii"))
    print("D... stream ok1", base64string)
    request.add_header("Authorization", "Basic %s" % base64string.decode("utf-8"))

    # request.add_header("Authorization", "Basic %s" % base64string)
    print("D... stream ok2 - request.urlopen (disp)")
    ok = False
    try:
        stream = urllib.request.urlopen(
            request, timeout=3
        )  # timeout to 7 from 5 sec.
        ok = True
        filesource = True
        print("D... stream ok3")
    except urllib.error.HTTPError as e:
        print("Server Offline1? Http error", e)
        print(videodev)
        # do stuff here
    except urllib.error.URLError as e:
        print("Server Offline2?  URL error", e)
        print(videodev)
        # do stuff here
    except:
        ok = False
        stream = None
        print("X.... Timeouted on URLOPEN")

    return stream, u, p







def setupsave(resolution=(640, 480),  videodev="xxx" ):
    sname = "rec"
    sname = videodev
    sname = sname.replace("http", "")
    sname = sname.replace("//", "")
    sname = sname.replace(":", "")
    sname = sname.replace("5000/video", "")
    sname = sname.replace("8000/video", "")

    sfilenamea = dt.datetime.now().strftime("%Y%m%d_%H%M%S")
    sme = socket.gethostname()
    sfilenamea = f"{sme}_{sname}_{sfilenamea}"
    #
    # DEFINE EXTENSION : avi+xvid     mov  mp4 mpg
    #   avi+x264
    #
    # sfilenamea = f"{sfilenamea}.mp4"
    # sfilenamea = f"{sfilenamea}.mp4"
    sfilenamea = f"{sfilenamea}.mkv"

    # XVID-works     LAGS   X264    MP4V? mp4v-new,ubu22    IYUV-huge
    CODEC = "mp4v"

    dir2create = os.path.expanduser("~/DATA/")
    if not os.path.isdir(os.path.expanduser(dir2create)):
        print(f"D... trying to create directory {dir2create} for saving")
        # result = False
        os.mkdir(os.path.expanduser(dir2create))

    sfilenamea = os.path.expanduser("~/DATA/" + sfilenamea)
    # codec
    # sfourcc = cv2.VideoWriter_fourcc(*'XVID')
    # ### sfourcc = cv2.VideoWriter_fourcc(*'LAGS') # lossless NOT mkv,avi
    # sfourcc = cv2.VideoWriter_fourcc(*'X264') # +avi small NOTubu22
    # sfourcc = cv2.VideoWriter_fourcc(*'mp4v') # works on UBU22 clean inst
    # ####sfourcc = cv2.VideoWriter_fourcc(*'MP4V')
    # ####sfourcc = cv2.VideoWriter_fourcc(*'IYUV') # huge + avi

    sfourcc = cv2.VideoWriter_fourcc(*f"{CODEC}")  #

    #  25 FPS
    saviout = cv2.VideoWriter(sfilenamea, sfourcc, 25.0, resolution)
    for i in range(4):
        print(f"SAVE={sfilenamea} ---  FOURCC={CODEC}")
    return sfilenamea, saviout



#!/usr/bin/env python3

from fire import Fire
# run this program on the Mac to display image streams from multiple RPis
import cv2
import imagezmq

import datetime as dt
import os
import sys

import notifator.telegram as telegram

import threading
import time
import numpy as np

import datetime as dt
import socket

allimages={}


results = [None]*2

MYSTART = dt.datetime.now()
OPENEDWR = {}

def receive_image(image_hub, results):
    rpi_name, image = image_hub.recv_image( )
    #return rpi_name, image
    results[0] = rpi_name
    results[1] = image



def setupsave( rpi_name,  resolution = (640,480) ):
    global MYSTART, OPENEDWR
    sme = socket.gethostname()

    # identifiable name
    sfilenamea = f'~/DATA/ZMQ_{sme}_{rpi_name}_{MYSTART.strftime("%Y%m%d_%H%M%S")}.mkv'
    sfilenamea = f'~/DATA/ZMQ_{sme}_{rpi_name}_{MYSTART.strftime("%Y%m%d_%H%M%S")}.avi'


    dir2create = os.path.expanduser("~/DATA/")
    if not os.path.isdir( os.path.expanduser(dir2create )):
        print(f"D... trying to create directory {dir2create} for saving")
        #result = False
        os.mkdir( os.path.expanduser(dir2create ))

    sfilenamea = os.path.expanduser(sfilenamea)
    if os.path.exists( sfilenamea ):
        return sfilenamea,OPENEDWR[sfilenamea]

    # codec
    #sfourcc = cv2.VideoWriter_fourcc(*'XVID')
    ### sfourcc = cv2.VideoWriter_fourcc(*'LAGS') # lossless NOT mkv,avi..,
    #sfourcc = cv2.VideoWriter_fourcc(*'X264') # +avi small
    #sfourcc = cv2.VideoWriter_fourcc(*'MP4V')

    sfourcc = cv2.VideoWriter_fourcc(*'IYUV') # huge + avi
    #  25 FPS
    saviout = cv2.VideoWriter( sfilenamea,sfourcc,1.0, resolution )
    OPENEDWR[sfilenamea] = saviout
    return sfilenamea,saviout





def main( display=False,  test= False):
    global MYSTART, OPENEDWR
    if test:
        print("test")
        cap = cv2.VideoCapture(0)
        ret, frame = cap.read()
        cv2.imshow('WebCam', frame)
        print( type(frame) )

        if type(frame) == np.ndarray:
            print("i... NUMPY ARRAY ####################################")
        time.sleep(1)

        telegram.bot_send("ALERT",f"test izmq {dt.datetime.now().strftime('%a %H:%M:%S')}", frame)

        time.sleep(2)
        cap.release()
        cv2.destroyAllWindows()

        sys.exit(0)
        # print("test")
        # telegram.bot_send("Login",f"test izmq {dt.datetime.now().strftime('%a %H%M%S')}")
        # sys.exit(0)





    image_hub = imagezmq.ImageHub()
    while True:  # show streamed images until Ctrl-C
        print("i... waiting receive..." )

        # i cannot use threading, it is returning parameters

        #x = threading.Thread( target=image_hub.recv_image , args=() )
        x = threading.Thread( target = receive_image , args=(image_hub, results) )
        x.start()
        while x.is_alive():
            #print("x", flush=True, end="")
            print(f"i... {dt.datetime.now().strftime('%a %H:%M:%S')} display={display}        ",  end="\r")
            time.sleep(0.05)


            #--------------------------------display section --
            if display:
                #print( allimages )
                for k,i in allimages.items():
                    #print(f"   key = {k}")
                    cv2.imshow( k, i) # 1 window for each RPi
                cv2.waitKey(1)


        x.join()
        #print(results)
        rpi_name = results[0]
        image = results[1]

        #rpi_name, image = image_hub.recv_image( )


        print(f"\ni... received {rpi_name}")
        rpi_name = f"ZMQ_Camera_{rpi_name}"
        # it appeared  ZMQ_gigajm_ZMQ_Camera_rpi4b456_20
        rpi_name = f"{rpi_name}"

        allimages[rpi_name] = image # if I have several incomming images...


        #fname =  f'~/DATA/{rpi_name}_{dt.datetime.now().strftime("%Y%m%d_%H%M%S")}.jpg'
        #fname = os.path.expanduser( fname )
        #print(fname)
        #cv2.imwrite( fname ,  image )

        sfilenamea,saviout = setupsave(rpi_name)
        print(f"i... writing image to  {sfilenamea} ")
        #saviout.write( image ) # I USED 2 images at once......
        saviout.write( image )

        if (dt.datetime.now()-MYSTART).total_seconds()>24*3600:
            MYSTART = dt.datetime.now() # .strftime("%Y%m%d_%H%M%S")
            for (i,v) in OPENEDWR.items():
                ## NO : 'NoneType' object has no attribute 'release'
                if not v is None:
                    v.release()
                OPENEDWR[i]=None




        #cv2.waitKey(1)
        image_hub.send_reply(b'OK')

if __name__=="__main__":
    Fire(main)




#!/usr/bin/env python3

from fire import Fire
from flashcam.version import __version__
import os
import cv2
from PIL import ImageFont, ImageDraw, Image
import numpy as np


#  Code:  ttfname, FSize(pt), KERN, FHieght, FWidth
#Fheight=>shiftUP txt by it
# default  10,0,10,5
fonts_available = {
    "s8":["small_pixel.ttf",8,0,7,5],
    "s6":["small_pixel.ttf",6,0,6,5],
    "pr":["prstartk.ttf",6,-1,6,5],
    "pr4":["prstartk.ttf",10,-1,10,9],
    "fj":["PixelFJVerdana12pt.ttf",4,0,6,5],
    "fj6":["PixelFJVerdana12pt.ttf",6,0,12,6],
    "re":["retganon.ttf",12,0,9,4],
    "p8":["pixelFJ8pt1__.TTF",8,0,6,5] ,
    "p7":["pixelFJ8pt1__.TTF",7,0,6,5] ,
    "ty":["TypographerFraktur-Medium.ttf",14,0,16,7],
    "v1":["visitor1.ttf",10,0,7,5],
    "v14":["visitor1.ttf",14,0,10,7 ],
    "v2":["visitor2.ttf",12,0,6,5],
    "co":["coders_crux.ttf",16,-1,7,5],
#    "ji":["JikiTourai.ttf",10,0,10,6],
#    "ca":["Caracas.ttf",10,0,10,7],
#    "sa":["SAIBA-45.ttf",12,0,10,7],
#    "ni":["NimbusMono-Regular.otf",12,0,9,7],
#    "nib":["NimbusMono-Bold.otf",12,0,9,7],
#    "li":["LiberationMono-Regular.ttf",12,0,10,7],
#    "cu":["COURIERO.TTF",10,0,8,7],
    "di":["digital-7.mono.ttf",10,0,8,5],
    "di99":["digital-7.mono.ttf",120,0,170,15],
    "vt":["VT323-Regular.ttf",12,-1,6,5],
    "vt4":["VT323-Regular.ttf",14,-1,13,5],
    "utfu":[ "Uni_Sans_Thin.otf", 24, 0, 24, 8 ],  # uppercase thin
    "utfo":[ "OpenSans-Regular.ttf", 24, 0, 30, 12 ],
    "utfg":[ "good_times_rg.otf", 24, 0, 24, 8 ], # uppercase - broad
    "utfp":[ "pricedown_bl.otf", 24, 0, 24, 8 ], # uppercase BoldSignature
    "utfh":[ "hemi_head_bd_it.otf", 24, -2, 24, 14 ]
}

LAST_FONT = None

def set_def_font( font ):
    global LAST_FONT
    if font is None: LAST_FONT = "p8"
    if font in fonts_available.keys():
        LAST_FONT = font
    else:
        LAST_FONT = "p8"


def get_def_font():
    global LAST_FONT
    return LAST_FONT


def get_f_width(font = None):
    global LAST_FONT
    mfont = font
    if mfont is None: mfont = LAST_FONT
    if mfont is None: return 5
    if mfont in fonts_available.keys():
        #LAST_FONT = mfont
        return fonts_available[mfont][4]
    else:
        return 5

def get_f_height(font = None):
    global LAST_FONT
    mfont = font
    if mfont is None: mfont = LAST_FONT
    if mfont is None: return 25
    if mfont in fonts_available.keys():
        #LAST_FONT = mfont
        return fonts_available[mfont][3]
    else:
        return 25

def iprint( frame, drtext, font, position , color_bgra=(100,0,0,0)  ):
    global LAST_FONT
    mfont = font
    if mfont is None: mfont = LAST_FONT
    if mfont is None: return frame
    if mfont in fonts_available.keys():
        #print("D.... FONT ACCEPTED:", mfont )
        #print("D.... FONT ACCEPTED:", fonts_available[mfont][0] )
        #LAST_FONT = mfont
        # prepare font
        fname = fonts_available[mfont][0]
        fsize = fonts_available[mfont][1]
        fkern = fonts_available[mfont][2]
        fheig = fonts_available[mfont][3]
        fontpath = os.path.expanduser(f"~/.config/flashcam/{fname}")
        font = ImageFont.truetype(fontpath, fsize)
        # draw
        img_pil = Image.fromarray(frame)
        draw = ImageDraw.Draw(img_pil )

        letter_spacing = fkern
        x,y=position
        # Draw each character with custom spacing
        for char in drtext:

            draw.text((x, y-fheig), char, fill=color_bgra, font=font, align="left")
            char_width, _ = draw.textsize(char, font=font)
            x += char_width + letter_spacing
        frame = np.array(img_pil)
    return frame



def dial(frame, position, radius=50, thickness=10, color=(0, 255, 0), percentage=100):
    """
    AI generated: too many iterations
use with text:
     FO="v14"
     frame = iprint( frame,
             str(round(100*overtextalpha/255)),
             FO,
             position = (post[0]-2*get_f_width(FO)//2,
             post[1] +get_f_height(FO)//2 ),
             color_bgra= (0,255, 0, 255-int(overtextalpha))
            )

    """
    # Convert the image to a PIL Image
    pil_img = Image.fromarray(frame)
    draw = ImageDraw.Draw(pil_img)

    # Calculate the bounding box of the circle
    left_up_point = (position[0] - radius, position[1] - radius)
    right_down_point = (position[0] + radius, position[1] + radius)

    # Calculate the start angle based on the percentage for clockwise drawing
    end_angle = 270  # Ending at the top (north)
    start_angle = end_angle - (360 * (percentage / 100.0))

    # Draw the partial circle (arc)
    draw.arc([left_up_point, right_down_point], start=start_angle, end=end_angle, fill=color, width=thickness)

    # Draw a thin full circle as a guide
    guide_thickness = 1  # Set the thickness of the guide circle
    draw.ellipse([left_up_point, right_down_point], outline=color, width=guide_thickness)

    # Convert back to OpenCV image


    frame = np.array(pil_img)
    FO="v14"
    frame = iprint( frame,
                    str( round(percentage)),
                    FO,
                    position = (position[0]-2*get_f_width(FO)//2,
                                position[1] +get_f_height(FO)//2 ),
                    color_bgra= ( color[0],color[1], color[2] )
                   )
    return frame


def signal_strength(image, position, size=50, color=(0, 255, 0), percentage=100):
    """
    by AI:
    """
    ratio = 2.33
    #print( type(position), position[0], position[1] )
    # Calculate the vertices of the triangle
    bottom_left = position
    bottom_right = (position[0] + size, position[1])

    top_right = (position[0] + size, position[1] - int(size/ratio) )

    # Draw the hollow green triangle
    cv2.line(image, bottom_left, bottom_right, (0, 255, 0), 1)
    cv2.line(image, bottom_right, top_right, (0, 255, 0), 1)
    #cv2.line(image, top_right, bottom_left, (0, 255, 0), 1)

    # Calculate the fill level based on the percentage
    fill_level = int(size * (percentage / 100.0))

    # Draw the filled part of the triangle
    if fill_level > 0:
        pts = np.array([[bottom_left,
                         (bottom_left[0] + fill_level, bottom_left[1]),
                         (bottom_left[0] + fill_level, bottom_left[1] - int((fill_level / size) * int(size/ratio)))]], np.int32)
        cv2.fillPoly(image, [pts], (0, 255, 0))

    FO="v14"
    image = iprint( image,
                    str( round(percentage)),
                    FO,
                    position = (position[0] -get_f_width(FO)//2,
                                position[1] -get_f_height(FO) ),
                    color_bgra= ( color[0],color[1], color[2] )
                   )
    return image


def text_box(image, position, box_size=(80,120), color=(0,255,0),split_height=20,  title="Title", values = "rate 123.2 uSv\n tot 220 mSv"):
    # Calculate the corners of the box
    top_left = position
    FO="v14"
    maxw=0
    for i in values.split("\n"):
        a = get_f_width(FO)*(len(str(i))+2)
        if a>maxw: maxw=a

    #xbox_size[1] = maxw
    xbox_size = [ maxw  , get_f_height(FO)* (len(values.split("\n"))+3)  ]

    bottom_left = (position[0], position[1] + xbox_size[1])
    top_right = (position[0] + xbox_size[0], position[1])
    bottom_right = (position[0] + xbox_size[0], position[1] + xbox_size[1])

    # Draw the rectangle for the text box in green
    cv2.rectangle(image, top_left, bottom_right, (0, 255, 0), 1)

    # Draw the horizontal split line in green
    split_line_start = (position[0], position[1] + split_height)
    split_line_end = (position[0] + xbox_size[0], position[1] + split_height)
    cv2.line(image, split_line_start, split_line_end, (0, 255, 0), 1)

    FO = "v14"
    image = iprint( image,
                    str( title),
                    FO,
                    position = (position[0]  +get_f_width(FO),
                                position[1] + 1.2*get_f_height(FO) ),
                    color_bgra= ( color[0],color[1], color[2] )
                   )
    nsk=0
    for i in values.split("\n"):
        image = iprint( image,
                        str( i),
                        FO,
                        position = (position[0] +get_f_width(FO),
                                    position[1] +(nsk+3)* get_f_height(FO) ),
                        color_bgra= ( color[0],color[1], color[2] )
                       )
        nsk+=1


    return image

def main():
    print()

if __name__=="__main__":
    Fire(main)




import numpy as np

import cv2
import math




def matrix_undistort(frame):
    """
    if apply_distortion...
    not used now
    """
    #dist = np.zeros((5,1))
    dist = np.array([[ -8e-7, 4e-11, 0.0, 0.0, 0.0]])

    #mtx = np.eye(3)
    scale = 1.0
    mtx = np.array([[scale, 0, h/2], [0, scale, w/2], [0, 0, 1]])

    h, w = frame.shape[:2]

    newcameramtx, roi = cv2.getOptimalNewCameraMatrix(mtx, dist, (w,h), 1, (w,h))
    # undistort
    dst = cv2.undistort(frame, mtx, dist, None, newcameramtx)
    ### crop the image
    ##x, y, w, h = roi
    return dst
    #frame = dst#[y:y+h, x:x+w]




def img_estim(img, thrshld=127):
    """
    not used
    """
    res = np.mean(img)
    return res
    is_light = res > thrshld
    return "light" if is_light else "dark"
    # 40 -> 2.2



def adjust_gamma(image, gamma=1.0):
    """
    local gamma mapped to d shift-d  ctrl-d
    """
    # build a lookup table mapping the pixel values [0, 255] to
    # their adjusted gamma values
    invGamma = 1.0 / gamma
    table = np.array(
        [((i / 255.0) ** invGamma) * 255 for i in np.arange(0, 256)]
    ).astype("uint8")
    # apply gamma correction using the lookup table
    return cv2.LUT(image, table)





def rotate_image(image, angle):
    if angle is None:     return image
    if abs(angle)<0.1:     return image
    image_center = tuple(np.array(image.shape[1::-1]) / 2)
    # print( "rotate", image_center, angle )
    rot_mat = cv2.getRotationMatrix2D(image_center, angle, 1.0)
    #print(rot_mat)
    result = cv2.warpAffine(image, rot_mat, image.shape[1::-1], flags=cv2.INTER_LINEAR)
    #print("rotated by ", angle)
    return result




# -------------------- DISPLAY MULTITEXT ---------------------
def disp_mutext(lena, wrapped_text_o):
    lena = np.array(lena)  # I turn to np.array
    lena = lena / 255

    size = lena.shape
    wrapped_text = wrapped_text_o.split("\n")

    img = np.zeros(size, dtype="uint8")
    img = img + 0.0  # +0.9 # whitish
    # print("npzero",img.shape, img)

    height, width, channel = img.shape

    # never used text_img = np.ones((height, width))
    # print(text_img.shape)
    font = cv2.FONT_HERSHEY_SIMPLEX

    x, y = 10, 40
    font_size = 0.4
    font_thickness = 1

    i = 0

    textsize1 = 1
    textsize0 = 1

    for line in wrapped_text:
        textsize = cv2.getTextSize(line, font, font_size, font_thickness)[0]
        if textsize[0] > textsize0:
            textsize0 = textsize[0]
        if textsize[1] > textsize1:
            textsize1 = textsize[1]

    #    wrapped_text.append( " "*textsize1+"[OK]" )
    # print(    wrapped_text )

    for line in wrapped_text:
        textsize = cv2.getTextSize(line, font, font_size, font_thickness)[0]
        if textsize[0] > textsize0:
            textsize0 = textsize[0]
        if textsize[1] > textsize1:
            textsize1 = textsize[1]
    # ----- after finding the text sizes; define gap

    gap = textsize1 + 6

    nlines = len(wrapped_text)
    offx = 0 + int((img.shape[1] - textsize0) / 2)
    offy = 0 + int((img.shape[0] - gap * (nlines - 1)) / 2)

    pad = 10
    start_point = (offx - pad, offy - pad - textsize1)
    start_point2 = (offx - pad, offy - pad - textsize1 + int(pad / 2))
    end_point = (pad + offx + textsize0, offy + gap * len(wrapped_text))
    end_point2 = (pad + offx + textsize0, offy + gap * len(wrapped_text) - int(pad / 2))

    # i guess this is gray ...
    img = cv2.rectangle(img, start_point, end_point, (0.3, 0.3, 0.3), -1)
    # this is a frame and  topbar
    img = cv2.rectangle(img, start_point, end_point, (-1, -1, -1), 1)  # trick
    img = cv2.rectangle(img, start_point2, end_point2, (-1, -1, -1), 1)

    for line in wrapped_text:
        textsize = cv2.getTextSize(line, font, font_size, font_thickness)[0]
        # print(textsize)
        # gap = textsize[1] + 5
        # gap = textsize1 # gap define earlier

        y = int((img.shape[0] + textsize[1]) / 2) + i * gap
        x = 10  # for center alignment => int((img.shape[1] - textsize[0]) / 2)
        x = offx
        y = offy + i * gap

        cv2.putText(
            img,
            line,
            (x, y),
            font,
            font_size,
            #                (255,255,255),
            (-1, -1, -1),  # BIG TRICK
            font_thickness,
            lineType=cv2.LINE_AA,
        )
        i += 1

    # -------------------- howto merge with frame........
    img = lena - img

    # print("min", img.min()  , "max=",img.max()  )
    img = np.clip(img, 0.0, 1.0, None)  # for dark img values, clip to 0
    # img = img / img.max()
    # print("min", img.min()  , "max=",img.max()  )

    img = img * 255  # i dont know how to make int
    img = img.astype(np.uint8)  # this makes negative to positive
    # print(img)
    return img



# ===================================== big mess


def make_measurements(frame, measure_fov, zoomme, measure, tracker1, tracker_list, cross):
    """
    fov...   measure ... parameters for measurement
    tracker1,_list. ... for speed measurement
    cross ... for movement of the point
    """
    h, w = frame.shape[0], frame.shape[1]
    # measure_fov = 110.5 # notebook

    # approximative, not precise... 5%
    radians_per_pixel = (measure_fov / 180 * math.pi) / w

    # rad per pix * unit distance * better
    # worked with zoom cameras????
    radians_per_pixel2 = math.tan(measure_fov / 180 * math.pi / 2 / w)

    # works with my 101deg camera---------------- and 55deg builtin ntbcam
    #           ..... for 0.5 tan and alpha have 10% diff (tan vs. atan)
    # measure_fov is TOTAL ANGLE
    # radians_per_pixel2 =  math.atan(measure_fov /180*math.pi/(2*w) )*2
    radians_per_pixel2 = math.atan(measure_fov / 180 * math.pi / w)
    # print( radians_per_pixel2 ) # GOOD .... 0.00275

    # radians_per_pixel2 =

    radians_per_pixel2 /= zoomme  # process with zoom

    # print(f"RPPX {radians_per_pixel}  {radians_per_pixel2} ")

    # now arbitrarily define 1 meter..like.. 100px =>
    # alpha = 100*radians_per_pixel
    # b = 1m / math.tan( alpha )

    def get_order(dist=1.7):  # determine order that fits
        # list of marks on the ruler
        wide = 0.001
        notfulrng = 1.0 # notfulrng==0.8; I want full range now..
        while True:
            wide *= 10
            pixwid = math.atan(wide / dist) / radians_per_pixel2
            alpha = pixwid * radians_per_pixel2
            if pixwid > w / 2 * notfulrng:  # not full rANGE
                wide /= 10
                pixwid = math.atan(wide / dist) / radians_per_pixel2
                alpha = pixwid * radians_per_pixel2
                break
        order = wide

        row = []

        while True:
            wide += order
            pixwid = math.atan(wide / dist) / radians_per_pixel2
            alpha = pixwid * radians_per_pixel2
            if pixwid > w / 2 * notfulrng:  # not full rANGE:
                wide -= order
                pixwid = math.atan(wide / dist) / radians_per_pixel2
                # neverused alpha = pixwid * radians_per_pixel2
                break
            else:
                row.append(wide)
        # -----
        # -----
        row.append(order)
        row = sorted(row)
        row = row[::-1]  # revert - we want Big to small
        # print( "TICKS... max:", row[0]  )

        base = 10**math.floor( math.log10( row[0]) )
        if row[0]/base<4:
            base = 10**math.floor( math.log10( row[0]/4) )

        minors = np.arange( base, row[0], base)
        minors = minors[::-1]
        #print(minors)
        if len(row) <= 2:
            in0 = row[0] / 2
            row.append(in0)
            in1 =  row[0] / 4
            row.append(in1)
        #     in2 = row[0] / 4 *3
        #     in3 = row[0] / 5
        #     #in2 = row[-1]/10
        #     row.append(in2)
        #     # row.append( in2 )
            # print("   > ",row)
        return row,minors  # Big to small


    def one_mark(dist=1.7, wide=[1, 2], speed=0, dispnumb = True):
        #  wide ...  # Big to small
        # h,w = frame.shape[0], frame.shape[1]
        # pixel distance of halfwidth

        # alpha = pixwid * radians_per_pixel2
        # dist = wide/math.tan( alpha)
        # I need to calculate 1m
        level = 0
        # print("XXXXX",wide)
        for iwide in wide:

            # multiply and round to one digit..before pixel calc
            for ix in range(4):
                if iwide*10**ix > 1:
                    iwide = round( iwide* 10**ix)/10**ix
                    break

            pixwid = math.atan(iwide / dist) / radians_per_pixel2
            # neverused alpha = pixwid * radians_per_pixel2

            # print(f" {radians_per_pixel}radpp {pixwid}   {wide}m <- {dist} ")
            step = 0
            mX, mY = int(w / 2), int(h / 2)
            if not (cross is None):
                mX += cross[0]
                mY += cross[1]
            # here I addd the red cross position

            mY = mY + level * step

            yA, yB = mY, mY

            xA = mX
            xB = mX + int(pixwid)

            yC = mY - int(pixwid) # up

            color = (0, 255, 0)  # BGR
            color = (55, 0, 255)  # BGR same as the red cross
            colos = (255, 0, 55)  # BGR same as the red cross
            if level == 0:
                # line - horiznotal
                cv2.line(
                    frame,
                    (int(xA), int(yA)),
                    (int(xB), int(yB)),
                    color,
                    1,
                )
                cv2.line( # left
                    frame,
                    (int(xA), int(yA)),
                    (int(2*xA-xB), int(yB)),
                    color,
                    1,
                )


                cv2.line( # probably central  mark
                    frame,
                    (int(xA), int(yA + 8)),
                    (int(xA), int(yA - 8)),
                    color,
                    1,
                )

            # vert bars on horiz axis
            cv2.line(  # ticks right
                frame,
                (int(xB), int(yB + 3)),
                (int(xB), int(yB - 3)),
                color,
                1,
            )
            cv2.line( # ticks left
                frame,
                (int(2*xA-xB), int(yB + 3)),
                (int(2*xA-xB), int(yB - 3)),
                color,
                1,
            )

            if yC>0:
                # horz bars on vertic axis
                cv2.line( # up
                    frame,
                    (int(xA), int(yA)),
                    (int(xA), int(yC)),
                    color,
                    1,
                )
                cv2.line(
                    frame,
                    (int(xA-3), int(yC )),
                    (int(xA+3), int(yC )),
                    color,
                    1,
                )

            unit = "m"
            # --- check the biggest to set the unit [0] is biggest
            if wide[0] <= 0.001:
                iwide = round(iwide * 1000 * 1000) / 1000
                unit = "mm"
            elif wide[0] <= 0.01:
                iwide = round(iwide * 100 * 1000) / 1000
                unit = "cm"
            elif wide[0] <= 0.1:
                iwide = round(iwide * 100 * 100) / 100
                unit = "cm"
            elif wide[0] < 1:
                iwide = round(iwide * 10 *100) / 100
                unit = "dm"
            else:
                iwide = round(iwide * 10) / 10

            # only properly round whatever unit
            if iwide <= 0.001:
                iwide = round(iwide * 10000) / 10000
            elif iwide <= 0.01:
                iwide = round(iwide * 1000) / 1000
            elif iwide <= 0.1:
                iwide = round(iwide * 100) / 100
            elif iwide < 1:
                iwide = round(iwide * 10) / 10
            else:
                iwide = round(iwide)


            if level > 0:
                unit = ""  # no unit during the scale
            unit2 = "m"

            if str(iwide)[:2] == "0.":
                iwide = str(iwide)[1:]

            if dispnumb:
                # width on scale - scale values
                cv2.putText(
                    frame,
                    f"{iwide}",
                    (int(xB - 10), int(mY - 7)),  # little rightx a bit up y
                    cv2.FONT_HERSHEY_SIMPLEX,
                    0.35,
                    color,
                    1,
                )
                cv2.putText(
                    frame,
                    f"{iwide}",
                    (int(2*xA-xB - 10), int(mY - 7)),  # little rightx a bit up y
                    cv2.FONT_HERSHEY_SIMPLEX,
                    0.35,
                    color,
                    1,
                )

                cv2.putText(  # up
                    frame,
                    f"{iwide}{unit}",
                    (int(xA + 10), int(yC)),  # little rightx a bit up y
                    cv2.FONT_HERSHEY_SIMPLEX,
                    0.35,
                    color,
                    1,
                )



                if level==0 and unit!="":
                    cv2.putText(
                        frame,
                        f"  unit ... {unit}",
                        (
                            int(xA - 130),
                            int(mY - 5 -40),
                        ),  # little rightx a bit up y
                        cv2.FONT_HERSHEY_SIMPLEX,
                        0.35,
                        color,
                        1,
                    )


            if level >= 0:
                # distance - only at first mark
                cv2.putText(
                    frame,
                    f"  at {dist} {unit2}",
                    (
                        int(xA - 130),
                        int(mY + 5 -40),
                    ),  # little rightx a bit up y
                    cv2.FONT_HERSHEY_SIMPLEX,
                    0.35,
                    color,
                    1,
                )
                cv2.putText(
                    frame,
                    f"  FOV {measure_fov:.1f}deg",
                    (
                        int(xA - 130),
                        int(mY + 15 -40),
                    ),  # little rightx a bit up y
                    cv2.FONT_HERSHEY_SIMPLEX,
                    0.35,
                    color,
                    1,
                )
                # I add velocity
                if (level == 0) and not (tracker1 is None):
                    cv2.putText(
                        frame,
                        f"speed {speed:6.2f}m/s",
                        (
                            int(xB - 50),
                            int(mY + 15),
                        ),  # little rightx a bit up y
                        cv2.FONT_HERSHEY_SIMPLEX,
                        0.35,
                        color,
                        1,
                    )
            level += 1

    # main part of the ruler making---------------

    order,minorticks = get_order(dist=measure)
    # print(minorticks)



    # speed computation
    if not (tracker1 is None) and len(tracker_list) > 2:
        b = tracker_list[-1]
        try:
            v_from = -2
            a = tracker_list[v_from]
            while True:
                dt = (b[2] - a[2]).total_seconds()
                if dt > 1.0:
                    break
                else:
                    v_from -= 1
                if len(tracker_list) < abs(v_from):
                    break
                a = tracker_list[v_from]
        except:
            dt = 100000  # velocity 0
        c = ((b[0] - a[0]) ** 2 + (b[1] - a[1]) ** 2) ** 0.5
        v = c / dt * radians_per_pixel2
        v = round(100 * math.tan(v) * measure) / 100
        # print(f"i... speed = {v} m/s  {dt:.2}==dt " )
    else:
        v = 0
    # plot ruler
    one_mark(dist=measure, wide=minorticks, speed=v, dispnumb = False)
    one_mark(dist=measure, wide=order, speed=v)
    return frame



#!/usr/bin/env python3
from imutils.video import VideoStream
import cv2
import time
import datetime as dt
import socket
from random import randint

import numpy as np

import imutils
import datetime as dt


from flashcam.version import __version__
import os


# run this program on each RPi to send a labelled image stream
import socket
import time
from imutils.video import VideoStream
import imagezmq
import zmq # for timeouts??
import threading


import notifator.telegram as telegram

from console import fg

# i want also here for LAP
from PIL import ImageFont, ImageDraw, Image

from console import fg,bg
from flashcam.text_write import iprint, fonts_available, get_f_height, get_f_width, get_def_font, set_def_font

# host name for once
SOCKET_GETHOSTNAME =socket.gethostname()

# =========================== DEFAULT FONT =====================
# get_f_height("vt4")
# #get_f_height("ty")
# #get_f_height("fj6")
# #get_f_height("re")
# #get_f_height("v14")
# #get_f_height("co")
# #get_f_height("pr4")
# get_f_height("nib")
# #get_f_height("li")
# #get_f_height("cu")
# get_f_height("di")
# #get_f_height("s8")

set_def_font("di")

class Stream_Enhancer:
    """
    definitions of positions on the frame
    ROW COLUMN  - colors defined elsewhere
   4 columns MAX
    """

    # COORDINATES used with setbox( "text", senh.TIME  ) _________
    # --------------- TOP  0,x
    TIME  = (0,0) #  basic bar
    DISP  = (0,4) # gree   -  detect delta histo
    MODE  = (0,3) # red    - DM z T
    SUBBG = (0,5) # substract background  or melt FG?

    # ------------- LOW LINE
    expo  = (2,0)
    gain  = (2,1)
    gamma = (2,2)
    speedx= (2,3)
    speedy= (2,4)
    #--------------- NEVER when speedxy
    dm=  (2,4) #  it is veeery long, overlaps

    #--------------------- 2 is bottom
    lap = (1,0) # side
    # avoid 2,0
    avg = (3,0) # accum    2,0 is for LOWLINE
    blr = (4,0) # LOW DEEP
    trh = (5,0) # side
    hist= (6,0) # side
    rot = (7,0) # side  /otate in bin...
    # ------ more (with/for uni)
    avi = (8,0) #
    jpg = (9,0) #

    scale = (10,0) #? see
    xzoom = (11,0) #? see

    delta_frame = None
    detect_frame = None
    frame_area = 1

    frame_number = 0

    motion_detected = False
    motion_detected_start = dt.datetime.now()
    expire = 15 # expire motiondetected (earlier 2 sec)
    aviopened = False

    avi_last = dt.datetime.now() - dt.timedelta(seconds=90000)
    aviopened_laps = False
    avi_started = dt.datetime.now()  # I like to have one file per day.....

    # -- age of frame for speed
    camera_age = 0
    camera_start = dt.datetime.now()

    # -- subtraction of mask
    frame_bg = None
    sub_bg = False
    frame_fg = None
    mix_fg = False

    # background path
    BGFILE = os.path.expanduser('~/.config/flashcam/background.jpg')
    FGFILE = os.path.expanduser('~/.config/flashcam/foreground.jpg')

    zmqrpi_name = socket.gethostname()
    zmqtarget = None #IP is operated from real_camera
    zmqsender = None # object
    zmqlast = dt.datetime.now()

    jtelegram = False # connector from real_camera
    telegram_tripwire = 2 # send 2 sec later after the triger...
    telegramlast = dt.datetime.now()
    telegramtrigger = False ## either BOOL or DATETIME

    aviout = None # I have a problem to seee saving the images to DaTA

    # this is used with setbox( "text", senh.TIME  ) _________
    def RC_COLOR(self,row,col):
        if not self.frame_ok: return  # B  G R
        if (row,col) == self.TIME: return (200,0,0)
        if (row,col) == self.MODE: return (0,0,255)
        if (row,col) == self.DISP: return (0,205,0)
        if (row,col) == self.SUBBG: return (90,50,200)

        if (row,col) == self.lap: return (200,100,200) #
        if row==1 and col==1: return (150,150,150) #
        if row==1 and col==2: return (50,50,50) #
        if row==1 and col==3: return (120,120,120) # brown
        if row==1 and col==4: return (180,180,180) # orange

        if (row,col) == self.avg:  return (250,0,0) #
        if (row,col) == self.blr:  return (150,150,0) #
        if (row,col) == self.trh:  return (150,0,150) #
        if (row,col) == self.hist: return (0,120,150) # brown
        if (row,col) == self.dm:   return (50,100,250) # orange
        if (row,col) == self.rot:  return (150,120,150) #


        # left side
        if (row,col) == self.expo:  return (0,0,0) #
        if (row,col) == self.gain:  return (50,50,50) #
        if (row,col) == self.gamma: return (80,80,80) #

        if (row,col) == self.speedx: return (180,80,80) #
        if (row,col) == self.speedy: return (80,180,80) #
        if (row,col) == self.avi:    return (30,30,210) # FOR UNI only
        if (row,col) == self.jpg:    return (0,0,250)   # FOR UNI only
        if (row,col) == self.scale:  return (250,0,250)   # zoom scale
        if (row,col) == self.xzoom:  return (0,100,250)   # xperimental zoom

        t=[255,255,255]
        while (t[0]+t[1]+t[2]>400) or  (t[0]+t[1]+t[2]<200) :
            t =  (randint(5,255),randint(5,255),randint(5,255) )
        #print(t[0]+t[1]+t[2])
        return t


#---------------------------------- INIT ----------

    def __init__(self, resolution=(640,480) ):
#    def __init__(self, resolution=(320,240) ):
        if  type(resolution)==str:
            w,h = resolution.split("x")
            w=int(w)
            h=int(h)
            print(f"i... {fg.yellow} STR StreamEnhancere init; resol== {w}x{h} {fg.default}")
        else:
            w,h=resolution[1],resolution[0]
            w,h=resolution[0],resolution[1]
            print(f"i... {fg.yellow} TUP StreamEnhancere init; resol== {w}x{h} {fg.default}")
        print(f"i... {fg.yellow} StreamEnhancere init; resol== {resolution} {fg.default}")
        print(f"i... {fg.yellow} StreamEnhancere init; resol== {w} x {h}  {fg.default}")
        self.resolution = (h,w)
        self.posx_offs=5
        self.posy_offs=5
        self.frame_ok = False

        self.first_frame = True
        self.accubuffer = [] # accumulation for motion
        self.accumulate_number = 1 # just informative for laps


#----------------------------------------- RETURN THE FRAME ----BACK
    def get_frame(self,  typ=""):
        if typ=="delta":
            if not(self.delta_frame is None):
                return self.delta_frame

        if typ=="detect":
            if not(self.detect_frame is None):
                return self.detect_frame

        if typ=="histo":
            #if "histo_frame" in locals():
            if not(self.histo_frame is None):
                return self.histo_frame

        return self.frame


#---------------------------  INSERT FRAME TO THIS OBJECT ----
    def add_frame(self, image):
        # print("D... adding fframe")
        h,w = 480,640
        self.frame_ok = True
        try:
           h,w = image.shape[0],image.shape[1]
        except Exception as ex:
            print(f"X... add_frame exc: ",ex)
            pass

        # EARLIER I DID CONVERT THE SIZE.... NOW I DO NOT WANT
        # BUT it is initialized already with 1280x720 e.g.
        # I CHANGE BRUTALY THE RESOLUTION NOW
        if image.shape[1]!=self.resolution[0] or image.shape[0]!=self.resolution[1]:
            #print("X... res0 == ",image.shape[0], self.resolution[0])
            #print("X... res1 == ",image.shape[1], self.resolution[1])
            #self.frame = cv2.resize(image, (self.resolution[1], self.resolution[0]), interpolation=cv2.INTER_NEAREST)
            #print(f" SH>>{fg.red}{self.frame.shape}{fg.default} ", end = "" )
            self.frame = image
            self.resolution = (image.shape[1], image.shape[0] )
            # print(f" {fg.red}SH!size{fg.default} ", end = "" )
        else:
            self.frame = image
        self.frame_ok = True
        #except Exception as e:
        #    print("D... in add frame bad frame")
        #    self.frame_ok = False
        if not self.frame_ok:
            return False

        self.maxcol=5 # NORMALLY 4
        # self.maxcol=7 # with x y
        self.maxrow=14
        # #if fontScale == 0.5
        #if w==320:
        #    self.maxcol=2
        #    self.maxrow=5
        # #if fontScale is W/640 *0.5
        if w==320:
        #    self.maxcol=2
            self.maxrow=13

        if self.first_frame:
            print( "D... 1st frame len=",len(self.frame) )
            self.averageValue1 = np.float32(self.frame)
            self.first_frame = False


        return True


#---------------------------  INSERT FRAME from CAMERA TO THIS OBJECT ----
    def add_frame_from_cam(self):
        """

        """
        self.imstream = VideoStream(0).start()
        self.imstream.stream.set(3, self.resolution[0])
        self.imstream.stream.set(4, self.resolution[1])
        image = self.imstream.read()
        self.imstream.stop()
        return self.add_frame(image)

#--------------------------- imshow ---------------------
    def show_frame(self):
        if not self.frame_ok: return
        cv2.imshow("A", self.frame)
        wkey="a"
        while wkey!=ord("q"):
            wkey = cv2.waitKey(10)
        cv2.destroyAllWindows()

#--------------------------- imshow ---------------------
    def blimp_frame(self):
        if not self.frame_ok: return
        cv2.imshow("A", self.frame)
        cv2.waitKey(100)

    def get_font_params(self, txt, fname=None):
        """
        I think this is obsolete
        """
        if not self.frame_ok: return
        self.font       = cv2.FONT_HERSHEY_SIMPLEX
        self.font       = cv2.FONT_HERSHEY_SIMPLEX
        self.font       = cv2.FONT_HERSHEY_DUPLEX

        self.font = cv2.FONT_HERSHEY_COMPLEX_SMALL
        self.font = cv2.FONT_HERSHEY_PLAIN  # needs offset

        self.fontScale  = 0.85 * (self.frame.shape[1]/640) # 0.5 * (self.frame.shape[1]/640)
        self.lineType   = 1
        self.text_width, self.text_height = cv2.getTextSize(txt,self.font,self.fontScale,self.lineType)[0]

        if fname is None:
            self.text_width, self.text_height = len(txt)*get_f_width(), get_f_height()
        #self.text_width, self.text_height = 15, 15



#------------------------------------------------------PUT TEXT --------------
    def textbox(self,txt, pos=[5,5], bgcolor=(0,0,0), fgcolor=(255,255,255) , target = None):
        if not self.frame_ok: return

        if target is None:
            framekind = self.frame
        else:
            framekind = target


        self.get_font_params(txt )
        #pos[1]+=self.text_height #

        # letwidth = 25*4*fontScale
        #        txt=txt+" w"+str(self.text_width)+" h"+str(self.text_height)+" @ "+str(pos[0])+","+str(pos[1])
        # top left 0 0
        beginCorner=( pos[0]-2                , pos[1]- self.text_height-1 )
        endCorner  =( pos[0]+self.text_width+1, pos[1]+1   )

        cv2.rectangle(framekind, beginCorner, endCorner, bgcolor, cv2.FILLED)

        self.frame = iprint( framekind, txt, get_def_font() , tuple(pos),  color_bgra=(255,255,255,0)  )
        # cv2.putText(framekind, txt,
        #             tuple(pos),
        #             self.font,
        #             self.fontScale,
        #             fgcolor,
        #             self.lineType)



#    def timemark(self):
#        # reserve 320
#        self.get_font_params(txt)
#        self.textbox(,  [self,5] )

#----------------------------------------------- DEFINE GRID HERE-------
    def setbox(self, txt, rowcol, side="" , target = None, grayed = False, kompr=None):
        """
        uses textbox, pots on positions....
        """

        if target is None:
            frametgt = self.frame
        else:
            frametgt = target

        row,col=list(rowcol)
        if not self.frame_ok: return
        stretchy = 2.2
        stretchx = 1.3 # this also compensates center-left align?
        pos=[ self.posx_offs, self.posy_offs ]
        self.get_font_params(" WEB 2.3, ")
        WIDTH = self.text_width
        self.get_font_params(txt )

        if side=="left":  col=0
        if side=="right":  col=-1
        pos[0] = self.posx_offs + int(col*WIDTH*stretchx)
        pos[1] = self.posy_offs + self.text_height + int(row*stretchy*self.text_height)
        pos[1] = 2+ self.text_height + int(row*stretchy*self.text_height)


        # Just columns
        if col>self.maxcol:
            print("X... only {} columns allowed".format(self.maxcol) )
            return

        # manually I set leftside
        # if (row==1)and(col==0):  side="left"
        # if (row==2)and(col==0):  side="left" # low deck
        if (row==3)and(col==0):  side="left"
        if (row==4)and(col==0):  side="left"
        if (row==5)and(col==0):  side="left"

        if (row==6)and(col==0):  side="left"
        if (row==7)and(col==0):  side="left"
        if (row==8)and(col==0):  side="left"
        if (row==9)and(col==0):  side="left"

        if (row==10)and(col==0):  side="left"
        if (row==11)and(col==0):  side="left" # xzoom

        if side=="" and row>2:
            print("X... SENH: Too many rows or undef.row ")
            return

        #------sides
        if side=="left" or side=="right":
            #print("D.. SIDES", self.maxrow, self.maxcol)
            if row>self.maxrow:
                print("X... {} rows {} cols allowed".format(self.maxrow,self.maxcol) )
                return
            #-----OK i go for it---
            #txt=txt[0:3] # i dont restrict...
            #print(txt)
            self.get_font_params("txt")
            WIDTH = self.text_width
            # ----- i dont remember why this ADDITION...
            #pos[1]+= int(2*stretchy*self.text_height)
            #---- right side
            if side=="right":
                pos[0] = self.frame.shape[1] - self.posx_offs - int(stretchx*WIDTH)


        else: # --------------------------- TimeStamp and G/R circle automatic HERE
            if row==0:
                if (col==0):
                    NOW = dt.datetime.now()
                    date = NOW.strftime("__%H:%M:%S %Y/%m/%d")
                    me = SOCKET_GETHOSTNAME #socket.gethostname()
                    txt = f"{date} {me} {__version__}"
                    # print("X... timemark requested")
                    self.textbox(txt, pos )

                    if kompr is not None:
                        # jpgkompr
                        self.get_font_params("99")
                        WIDTH = self.text_width
                        xmax = int(self.frame.shape[1] - 0.*self.posx_offs - int(stretchx*WIDTH) )
                        self.textbox( str(kompr), ( xmax,pos[1]) )

                    # RED GREEN CIRCLE:
                    #cirdiam = int(self.fontScale*10) # int(self.fontScale*16)
                    # not really # cirdiam = self.text_height + int(row*stretchy*self.text_height)
                    cirdiam = 1.2*get_f_height()/2
                    cirdiam = round(cirdiam)
                    if NOW.second % 2 == 0:
                        #  fontsize 0.5=> 10   or 0.3 for RPI 320x240
                        cv2.circle( self.frame, (cirdiam,cirdiam), cirdiam, (0,0,255), -1 )
                        # pass
                    else:
                        cv2.circle( self.frame, (cirdiam,cirdiam), cirdiam, (0,255,0), -1 )

                    return
                if col<3:
                    return


            #if row==2:
                #pos[1] = self.posy_offs + int(stretchy*self.text_height)
            if row==2:
                pos[1] = self.frame.shape[0] - int(1.5*self.posy_offs)

        #print(pos)
        bgcolor =  self.RC_COLOR(row,col)
        if grayed: bgcolor = (120,120,120)
        self.textbox(txt, pos , bgcolor= bgcolor , target = frametgt)

    #-------------------------------------------- ENDOF LABELS----------


    def rotate180(self, angle):
        #print(f" rotate {angle}")
        if angle == 0:
            return
        elif angle == 180:
            self.frame = cv2.rotate(self.frame, cv2.ROTATE_180)
        else:
            self.frame = imutils.rotate(self.frame, angle)

    def save_background(self):
        cv2.imwrite( self.BGFILE , self.frame ) # nicely saved
        self.frame_bg = self.frame.copy() # else artifacts remain
        # frame_bg = self.frame.copy()
        # frame_bg2 = cv2.imencode('.jpg', frame_bg)[1].tobytes()
        # save_bg = False

    def save_foreground(self):
        cv2.imwrite( self.FGFILE , self.frame ) # nicely saved
        self.frame_fg = self.frame.copy() # else artifacts remain
        # frame_bg = self.frame.copy()
        # frame_bg2 = cv2.imencode('.jpg', frame_bg)[1].tobytes()
        # save_bg = False

    def subtract(self):
        if self.frame_bg is None:
            if os.path.exists( self.BGFILE ):
                self.frame_bg = cv2.imread( self.BGFILE )
                print("OLD BACKGROUND IMAGE FOUND", self.frame_bg.shape)
            else:
                return
        frame1 = cv2.subtract(self.frame , self.frame_bg ) #, mask = frame_mask
        # frame 1 is difference...
        # frame1[:22,:]  = self.frame [:22,:]
        # frame1[-22:,:] = self.frame [-22:,:]
        self.frame = frame1



    def mix(self):
        if self.frame_fg is None:
            if os.path.exists( self.FGFILE ):
                self.frame_fg = cv2.imread( self.FGFILE )
                print("OLD FOREGROUND IMAGE FOUND", self.frame_fg.shape)
            else:
                return
        frame1 = 0.5*self.frame + 0.5*self.frame_fg
        #*imgs[ random.randint(0,len(imgs)-1) ]

        #frame1 = cv2.subtract(self.frame , self.frame_fg ) #, mask = frame_mask
        # frame 1 is difference...
        # frame1[:22,:]  = self.frame [:22,:]
        # frame1[-22:,:] = self.frame [-22:,:]
        self.frame = frame1



    def reset_camera_start(self):
        self.camera_start = dt.datetime.now()

    def translate(self, speedx = -0.35,  speedy = -0.015 ):
        #print(f"translate {speedx} {speedy}")
        if ( abs(speedx)>1 ) or ( abs(speedy)>1 ):
            self.camera_age = 1
        else:
            self.camera_age = (dt.datetime.now() - self.camera_start ).total_seconds()

        # --- for larger speeds
        dwidth = speedx * self.camera_age  #negative up
        dheight = speedy * self.camera_age  # positive down
        T = np.float32([[1, 0, dwidth], [0, 1, dheight]])
        shifted = cv2.warpAffine(self.frame, T, (self.frame.shape[1], self.frame.shape[0]))
        # print( T )
        self.frame = shifted



    def crosson(self,  dix, diy, color = "g", box_small = True, box_large = False):
        """
        two types  g (just cross, r boxed cross)
        """
        #if color == 'r': crotype = 'line'
        #
        RADIUS=63
        y = int(self.frame.shape[0]/2)
        x = int(self.frame.shape[1]/2)

        ix = x+dix
        iy = y+diy

        if color=="g":
            lcolor=(0,255,55)
        elif (color=="r"):
            lcolor=(55,0,255) #BGR
        else:
            lcolor=(0,255,55)

        crscal = 4
        crnx,crny = int(64/crscal),int(48/crscal)
        #if crotype == "box":
        midskip = crnx
        midskipy = crny
        #else:
        #    midskip = 7
        #    midskipy = 7


        i2=cv2.circle( self.frame, (ix,iy), RADIUS, lcolor, 1)
        i2=cv2.line(i2, (ix-RADIUS+midskip,iy), (ix-midskip,iy), lcolor, thickness=1, lineType=8)
        i2=cv2.line(i2, (ix+RADIUS-midskip,iy), (ix+midskip,iy), lcolor, thickness=1, lineType=8)

        i2=cv2.line(i2, (ix,iy-RADIUS+midskipy), (ix,iy-midskipy), lcolor, thickness=1, lineType=8)
        i2=cv2.line(i2, (ix,iy+RADIUS-midskipy), (ix,iy+midskipy), lcolor, thickness=1, lineType=8)

        # mid
        i2=cv2.line(i2, (ix,iy), (ix,iy), lcolor, thickness=1, lineType=8)

        #if crotype == "box":
        if box_small:
            #corners  #  position 0.5deg from 11 deg. OK
            crscal = 4 # normal original box
            crscal = 3.2 # normal original box
            crnx,crny = int(64/crscal),int(48/crscal)

            i2=cv2.line(i2, (ix-crnx,iy-crny), (ix+crnx,iy-crny), lcolor, thickness=1, lineType=8)
            i2=cv2.line(i2, (ix+crnx,iy-crny), (ix+crnx,iy+crny), lcolor, thickness=1, lineType=8)
            i2=cv2.line(i2, (ix+crnx,iy+crny), (ix-crnx,iy+crny), lcolor, thickness=1, lineType=8)
            i2=cv2.line(i2, (ix-crnx,iy+crny), (ix-crnx,iy-crny), lcolor, thickness=1, lineType=8)

        if box_large:
            #corners  #  position 0.5deg from 11 deg. OK
            crscal = 1.4 # normal original box
            crnx,crny = int(64/crscal),int(48/crscal)

            i2=cv2.line(i2, (ix-crnx,iy-crny), (ix+crnx,iy-crny), lcolor, thickness=1 )
            i2=cv2.line(i2, (ix+crnx,iy-crny), (ix+crnx,iy+crny), lcolor, thickness=1 )
            i2=cv2.line(i2, (ix+crnx,iy+crny), (ix-crnx,iy+crny), lcolor, thickness=1 )
            i2=cv2.line(i2, (ix-crnx,iy+crny), (ix-crnx,iy-crny), lcolor, thickness=1 )

        #return self.frame


    #---------- for motion detect i need a second processing line frame2
    # to keep the camera on

    def setblur(self, blur=0, dmblur = 0):
        if (blur>0) and (blur % 2 ==0):
            print("D... wrong gaussian value, incrementing by 1")
            blur+=1
        if blur>0:
            self.frame = cv2.GaussianBlur(self.frame, (blur, blur), 0)
        if (dmblur>0) and (dmblur % 2 ==0):
            dmblur+=1
        # if dmblur>0:
        #     grayB = cv2.cvtColor(self.frame, cv2.COLOR_BGR2GRAY)
        #     grayB = cv2.GaussianBlur(grayB, (dmblur, dmblur), 0)


    def setaccum(self, accumulate=0):
        self.accumulate_number = accumulate # info for laps
        if accumulate>1:
            cv2.accumulateWeighted(self.frame, self.averageValue1, 1/accumulate )
            #self.frame = self.averageValue1 all is white
            self.frame = cv2.convertScaleAbs(self.averageValue1)
        else: # reset
            self.averageValue1 = np.float32(self.frame)



    def chk_threshold(self, threshold=100):
        """
        compare threshold and eventually set DETECTED True
        """
        # NOT HERE; this is a timelock - only a new seq self.motion_detected = False #  Here the flag is risen
        motion = False
        self.detect_frame = self.frame.copy()
        self.frame_number+= 1

        #print( self.delta_frame)
        if not (self.delta_frame is None):
            # ------------- THIS IS TUNABLE FROM COMMANDLINE/CONFIG
            thresh = cv2.threshold(self.delta_frame, threshold, 255,
                                cv2.THRESH_BINARY)[1]
            # print(thresh)
            ok = False
            noz = 0 # non zero?
            try:
                noz = cv2.countNonZero(thresh)
                ok = True
            except:
                ok = False
            if not(ok):
                return

            # -----------------------------------
            thresh = cv2.dilate(thresh, None, iterations=3)
            cnts = cv2.findContours(thresh.copy(), cv2.RETR_EXTERNAL,
                                    cv2.CHAIN_APPROX_SIMPLE)
            cnts = imutils.grab_contours(cnts)

            totarea = 0
            largest = cnts # ?? trick later
            largesta = 0
            largest = []
            frame_area = self.delta_frame.shape[0]*self.delta_frame.shape[1]
            for c in cnts:
                dearea = cv2.contourArea(c)
                # print("    search {}".format( dearea ) )
                totarea+=dearea # total over threshold = totarea; X frame_area
                (x, y, w, h) = cv2.boundingRect(c)
                if (y+h<self.text_height*3) or (y>self.delta_frame.shape[0]-self.text_height*3):
                    continue
                cv2.rectangle(self.detect_frame, (x,y),(x+w,y+h),(0,255,0),1)
                if largesta<dearea:
                    largesta = dearea
                    largest = [c] # cheap trick, one instead of all

            # for c in cnts:
            largesta = int(largesta/frame_area*1000)/10 # recode for late


            # ------------------------------  MOTION FALSE --- AND CALCULATE
            motion = False
            # -----one in the list  for now.... if larger than 0.7 ==> i dont care...
            for c in largest: # THE LIST OF ONE ELEMENT - TRICK
                area = cv2.contourArea(c)
                # print("Largest: {}/{}  ... {:.2f}%".format(area, frame_area, area/frame_area*100) )
                # rectangl 1% IS FINE. I pUSH to 0.1%
                AREA_THR_MIN = 0.001
                AREA_THR_MAX = 0.7
                if (area/frame_area < AREA_THR_MIN) or (area/frame_area > AREA_THR_MAX): #minarea 0.5%
                    # go away for any smaller
                    continue
                # what is here means detection
                motion = True
                (x, y, w, h) = cv2.boundingRect(c)
                cv2.rectangle(self.detect_frame, (x,y),(x+w,y+h),(0,0,255),2)  # BGR
                area=w*h # of the largest box

                #  DM conditions in orange :  area/frame % / noz/frame % / cnts
                dtext = "{:.1f}%/{:.1f}%/{}".format(
                    area/frame_area*100, noz*100/frame_area, len(cnts) )

                # --- all with detection is done here.... detect and delta frames must exist then
                self.setbox( f"{dtext}", self.dm ,  target = self.detect_frame )
                self.setbox( f"{dtext}", self.dm ,  target = self.delta_frame )
                self.setbox( f"{dtext}", self.dm    )
                if "self.histo_frame" in locals():
                    self.setbox( f"{dtext}", self.dm ,  target = self.histo_frame )


            # all calculation done:
            # NEW MOTION:
            #if (motion) and (not self.motion_detected):
            # AnY Motion
            if motion:
                self.motion_detected_start = dt.datetime.now() # Set New Time
                #print("i... NEW motion detected", self.motion_detected_start, self.frame_number)

            if motion:
                self.motion_detected = True
                #print("i... Motion ON")  # comming here mns already TRUE

            # this was the protection not to save too long....
            #if ((dt.datetime.now() - self.motion_detected_start).total_seconds() > self.expire) :
            #    self.motion_detected = False # this kills
            #    # print("i... Motion OFF expired", self.expire, self.frame_number)


            # this IS now the forced recording .....
            if ((dt.datetime.now() - self.motion_detected_start).total_seconds() < self.expire) :
                #print(f"i...     {(dt.datetime.now() - self.motion_detected_start).total_seconds():.1f} s / {self.expire} s     " )
                proc = (dt.datetime.now() - self.motion_detected_start).total_seconds() / self.expire
                bar = 13 - int(proc*13)
                if not motion: self.setbox( " "*bar, self.dm    )
                self.motion_detected = True
            else:
                # it expired
                if not motion:
                    # and no motion is here
                    self.motion_detected = False
                    #print("i... Motion OFF", self.frame_number)
            #if len(largest)==0: # ----------- if no contour - set OFF
            #    self.motion_detected = False # this kills
                # print("i... Motion OFF", self.frame_number)




    def detmo(self,  dmaccumulate=0, dmblur = 0, threshold = 100 ):
        """
        if there is an accumulation - process it
        if there is NO detection, rewrite the buffer
        """
        if dmaccumulate>1:
            if (dmblur>0) and (dmblur % 2 ==0):
                dmblur+=1


            frame2 = None
            if dmblur==0:
                dmblur=1



            self.accubuffer.append(self.frame)
            while len(self.accubuffer)>dmaccumulate: # i compare against X frames back
                frame2=self.accubuffer.pop(0) # get more time distance

            if not (frame2 is None):
                # create version A
                if dmaccumulate!=0:
                    cv2.accumulateWeighted(frame2, self.averageValue1, 1/dmaccumulate, None)
                else:
                    cv2.accumulateWeighted(frame2, self.averageValue1, 1, None)

                grayA = cv2.cvtColor(cv2.convertScaleAbs(self.averageValue1), cv2.COLOR_BGR2GRAY)
                grayA = cv2.GaussianBlur(grayA, (dmblur, dmblur), 0)

                # create version B
                grayB = cv2.cvtColor(self.frame, cv2.COLOR_BGR2GRAY)
                grayB = cv2.GaussianBlur(grayB, (dmblur, dmblur), 0)
                # grayA = cv2.cvtColor(averageValue1, cv2.COLOR_BGR2GRAY)
                #----see the difference
                #frame = cv2.absdiff(grayA, grayB)
                self.delta_frame = cv2.absdiff(grayB, cv2.convertScaleAbs(grayA))
            else:
                print("i... None frame2")
                self.delta_frame = self.frame



# ------------------------------------------------------- HISTO ---------------------------
    def histo_medi(self):
        def find_med( arr ):
            # not transposed!
            bina = 0
            suma = 0
            for i in arr:
                #print(i)
                bina+=1
                suma+=i
                #print(suma)
                if suma>0.5:
                    return bina
            return len(arr)

        BINS = 256 # 128
        #bins = np.arange(BINS)
        bins = np.arange(BINS).reshape(1,BINS)
        framegray = cv2.cvtColor(self.frame, cv2.COLOR_BGR2GRAY)
        # images; channels; mask none; histSize [BINS]
        hist_gray = cv2.calcHist([framegray], [0], None, [BINS], [0, 256])

        #hist_gray = np.transpose( hist_gray) # array
        sumhist = hist_gray.sum()
        hist_gray = hist_gray/sumhist
        sumhist = hist_gray.sum()

        #print(hist_gray)
        #print(sumhist)
        med = np.median( hist_gray )
        med2= find_med( hist_gray)
        #print( f" sum = {sumhist:6.1f}    med={med:.5f}  bin={med2}" )
        return med2

        # cv2.normalize(hist_gray,hist_gray,0,255,cv2.NORM_MINMAX)
        #med =
        black = (hist_gray[0] + hist_gray[1] )/2 #1st two bins
        white = (hist_gray[-1]+hist_gray[-2] )/2 #last 2 bins

        mean = np.dot( bins, hist_gray )
        #print("MEAN",mean)
        mean = round(100*mean[0][0]/sumhist/BINS) # min is bin 1, max is bin BINS.

        # print(f"   {len(hist_gray)}  {hist_gray[0]}..{hist_gray[-1]}  sum={sumhist}  med={med},  mean={mean}  ")

        #print("MEAN",mean)
        #print( "LEN TYPE", len(mean),type(mean) )
        BW = ""
        if white>100: BW=BW+"W"
        elif white>0: BW=BW+"w"

        if black>100: BW=BW+"B"
        elif black>0: BW=BW+"b"

        #return mean
        return f"*{mean:.0f}{BW}"  #/255*100


    def histo_mean(self):

        BINS = 128
        #bins = np.arange(BINS)
        bins = np.arange(BINS).reshape(1,BINS)

        framegray = cv2.cvtColor(self.frame, cv2.COLOR_BGR2GRAY)

        # images; channels; mask none; histSize [BINS]
        hist_gray = cv2.calcHist([framegray], [0], None, [BINS], [0, 256])
        # cv2.normalize(hist_gray,hist_gray,0,255,cv2.NORM_MINMAX)

        #hist_gray = hist_gray/hist_gray.max() # norm 1
        #mean = 1

        #print(f"i... mimax mean  {hist_gray.min():4.1f},  {hist_gray.max():4.1f}, {hist_gray.mean():4.1f} ... {mean}" )
        #print( hist_gray[:9],  bins[:9]  , len(hist_gray), len(bins) , type(hist_gray), type(bins) )

        # strange problem WITH 046d:080f  Webcam C120
        #print(f"hg                     binslen={len(bins)}=1; BINS={BINS}  lenHG={len(hist_gray)}")
        sumhist = hist_gray.sum()
        #print(f"hg                     sumhist = {sumhist} all the same!")


        black = (hist_gray[0] + hist_gray[1] )/2
        white = (hist_gray[-1]+hist_gray[-2] )/2

        #print(f"hg              {max(hist_gray)} - {min(hist_gray)} wh{white} blk2{black}")
        #if sumhist<1:
        #    return 100 # I DONT KNOW WHAT IS THIS
        mean = np.dot( bins, hist_gray )
        #print("MEAN",mean)
        mean = round(100*mean[0][0]/sumhist/BINS) # min is bin 1, max is bin BINS.
        #print("MEAN",mean)
        #print( "LEN TYPE", len(mean),type(mean) )

        #cv2.normalize(hist_gray,hist_gray)
        #mean = hist_gray.mean()/BINS #/len(hist_gray)
        ###mean = mean.mean()/BINS
        # mean = hist_gray.mean() #/len(hist_gray)
        # min1 = hist_gray.min()
        # max1 = hist_gray.max() #/len(hist_gray)
        #print(f"i... .flattened gray    {mean}")
        #  {min1:.1f} {max1:.1f}

        BW = ""
        if white>100: BW=BW+"W"
        elif white>0: BW=BW+"w"

        if black>100: BW=BW+"B"
        elif black>0: BW=BW+"b"

        return mean
        return f"{mean:.0f}{BW}"  #/255*100




    def zoom(self, scale, x=None,y=None):
        #scale = zoom
        #get the webcam size

        if len(self.frame.shape)==3:
            height, width, channels = self.frame.shape
        else:
            height, width = self.frame.shape
        #prepare the crop
        centerX = int(height/2)
        centerY = int(width/2)

        #if not x is None:
        #    centerX+= x
        #if not y is None:
        #    centerY+= y
        if (x!=None) or (y!=None):
            dwidth = -x    #negative up
            dheight = -y   # positive down
            T = np.float32([[1, 0, dwidth], [0, 1, dheight]])
            self.frame = cv2.warpAffine(self.frame, T, (self.frame.shape[1], self.frame.shape[0]))

        # return
        radiusX,radiusY= int(height/2/scale),int(width/2/scale)

        minX,maxX=centerX-radiusX,centerX+radiusX
        minY,maxY=centerY-radiusY,centerY+radiusY
        #minX=max(minX,0)
        #minY=max(minY,0)
        #maxX=min(maxX,height)
        #maxY=min(maxY,width)

        #print(" cut x=",minX,maxX, "  and y=",minY,maxY )
        cropped = self.frame[minX:maxX, minY:maxY]
        #print("CENTER:",centerX, centerY,"#",  cropped.shape, scale,"x")
        self.frame = cv2.resize(cropped, (width, height), interpolation=cv2.INTER_NEAREST)


    def histo(self):

        h = self.frame.copy()
        BINS = 128
        bins = np.arange(BINS).reshape(BINS,1)
        color = [ (255,0,0),(0,255,0),(0,0,255) ]

        for ch, col in enumerate(color):
            hist_item = cv2.calcHist([self.frame],[ch],None,[BINS],[0,255])
            #cv2.normalize(hist_item,hist_item, 0, 255, cv2.NORM_MINMAX)
            hist_item = hist_item/hist_item.max()*255
            hist=np.int32(np.around(hist_item))
            pts = np.column_stack((bins*int(640/BINS),475-(hist*440/255).astype(int)))
            cv2.polylines(h,[pts],False,col)

        framegray = cv2.cvtColor(self.frame, cv2.COLOR_BGR2GRAY)
        hist_gray = cv2.calcHist([framegray], [0], None, [BINS], [0, 255])
        #cv2.normalize(hist_gray,hist_gray,0,255,cv2.NORM_MINMAX)
        hist_gray = hist_gray/hist_gray.max()*255

        hist=np.int32(np.around(hist_gray))
        pts = np.column_stack((bins*int(640/BINS),475-(hist*440/255).astype(int)))
        cv2.polylines(h,[pts], False,  [255,255,255], thickness= 2 )

        self.histo_frame = h


    def zmqsender_send_image(self):
        """
        just send, but catch exception....
        """
        try:
            self.zmqsender.send_image( self.zmqrpi_name, self.frame )
            print(f"i... zmq was sent to {self.zmqtarget}")
        except:
            print(f"i... zmq NOT sent to {self.zmqtarget}")
            self.zmqsender = None
        return


    def telegram_send_image(self, blocking = True):
        """
        send telegram from here... be sure:no flooding
        """
        # telegram
        TELEGRAMBLOCK = 300 # seconds
        if self.jtelegram and (os.path.exists(os.path.expanduser("~/.telegram.token"))):
            telegramnow = dt.datetime.now()
            # 5 minutes
            # protection against flooding
            if blocking==False or  (telegramnow-self.telegramlast).total_seconds()>TELEGRAMBLOCK:
                print(f"fCAMERA {self.zmqrpi_name} ")
                print(f"fCAMERA {self.zmqrpi_name} {telegramnow.strftime('%a %H:%M:%S')}")
                y = threading.Thread(target=telegram.bot_send, args=("ALERT", f"CAMERA {self.zmqrpi_name} {telegramnow.strftime('%a %H:%M:%S')}", self.frame))
                y.start()
                print("i.. telegram SENT============")
                self.telegramlast=telegramnow
            else:
                print(f"i.. telegram not allowed due to block-time ({TELEGRAMBLOCK} s); now=", (telegramnow-self.telegramlast).total_seconds())
        #else:
        #    print(f"X...  jtele=={self.jtelegram}, token")
        return



# --------------------------------------------------------save avi ------------
# ------------------------------------------------- also ZMQ
    def save_avi(self, seconds =  1, name = "" ,  basecamera_string= None ,
                 mycodec = "DIVX", frnum= -1,
                 container = "avi",
                 TIMEOUT_NEWFILE = 3600,
                 fps = 10.1):
        """
        seconds>0 for TIMELAPS  ;   0.001 intentional for every frame
         0 just return
        -1 with name="dm" for detectmotion
        frnum .. if >0 and seconds <0.1 => omit may omit accumulated
         ...
        """
        #print("i... save_avi", seconds, name)
        # -------------- one shot



        if seconds == 0:
            return
        elif seconds == -1: # <0 [ -1 now] means saving avi - motion detect or in uni...*******
            #
            #  NEGATIVE :  means saving all......
            #     USED ALSO and mainly FOR DETECTION
            #
            #
            # trigger is old now
            # if not(self.telegramtrigger)==bool:
            #     if self.telegramtrigger<=dt.datetime.now()+dt.timedelta(seconds=1):
            #         self.telegramtrigger = dt.datetime.now()+dt.timedelta(seconds=2)
            #         print(f"i... telegram preset to {self.telegramtrigger.strftime('%H:%M:%S')}")

            WH = (self.frame.shape[0],self.frame.shape[1]) # SHIIIIIT 2hours
            WH = (self.frame.shape[1],self.frame.shape[0])

            if not self.aviopened or  (dt.datetime.now()-self.avi_started).total_seconds()>TIMEOUT_NEWFILE:
                me = socket.gethostname()
                filename = dt.datetime.now().strftime("%Y%m%d_%H%M%S")
                filename = os.path.expanduser("~/DATA/"+f"{me}{name}_{filename}_{mycodec}.{container}")
                for i in range(3): print(filename)
                if (dt.datetime.now()-self.avi_started).total_seconds()>TIMEOUT_NEWFILE:
                    print("i... RLEASING VIDEOFILE ..... EVERY sec.==", TIMEOUT_NEWFILE)
                    self.aviout.release()
                    time.sleep(0.5)
                    print("i... RELEASING VIDEOFILE ..... EVERY sec.==", TIMEOUT_NEWFILE)
                    self.avi_started = dt.datetime.now()
                #
                fourcc = cv2.VideoWriter_fourcc(*mycodec)  #
                self.aviout = cv2.VideoWriter( filename , fourcc , fps, WH )
                self.aviopened = True
                self.aviout.write(self.frame)
                self.aviout.write(self.frame)
                self.aviout.write(self.frame)


                #--- what happens if not 640x480 and xvid?  XVID I ALWAYS USED.... sometimes too big compression
                #fourcc = cv2.VideoWriter_fourcc(*'XVID')

            #elif (self.aviopened) and ( (dt.datetime.now()-self.avi_started).total_seconds()>86400): # it is opened but too long, one day....
                #  print(self.aviout)

            #if self.aviout:
            #    #print("X")
            self.aviout.write(self.frame)
            # print("                                        ",WH, self.frame.shape, fps, end  = "\r")
            # self.aviout.release()
            # plan telegram send to future
            # bool means it is not set....
            #
            if self.jtelegram and type(self.telegramtrigger)==bool:
                self.telegramtrigger = dt.datetime.now()+dt.timedelta(seconds=self.telegram_tripwire) # TRIPWIRE 2 seconds
                print("i... telegramtrigger preset to +2 sec.",self.telegramtrigger.strftime("%H:%M:%S"))
            # MOMENT WHEN AVI IS SAVED===========> imagezmq, telegram...




            #print(self.zmqsender,self.zmqtarget)
            # zmqtarget - if not NONE not initialized
            if  (self.zmqsender is None) and not (self.zmqtarget is None):
                print(self.zmqsender,self.zmqtarget," INITIALIZING",f'tcp://{self.zmqtarget}:5555')

                self.zmqsender = imagezmq.ImageSender(connect_to=f'tcp://{self.zmqtarget}:5555')
                self.zmqsender.zmq_socket.setsockopt(zmq.LINGER, 0)  # prevents ZMQ hang on exit
                # NOTE: because of the way PyZMQ and imageZMQ are implemented, the
                #       timeout values specified must be integer constants, not variables.
                #       The timeout value is in milliseconds, e.g., 2000 = 2 seconds.
                self.zmqsender.zmq_socket.setsockopt(zmq.RCVTIMEO, 2000)  # set a receive timeout
                self.zmqsender.zmq_socket.setsockopt(zmq.SNDTIMEO, 2000)  # set a send timeout




            # if zmqtarget IP address is given:
            if not(self.zmqtarget is None) and not(self.zmqsender is None):
                #print(self.zmqsender,self.zmqtarget," action")
                zmqnow = dt.datetime.now()


                # imagezmq
                if (zmqnow-self.zmqlast).total_seconds()>1.5:
                    #x = threading.Thread(target=self.zmqsender.send_image, args=(self.zmqrpi_name, self.frame))
                    x = threading.Thread(target=self.zmqsender_send_image ) # i will try to catch exceptions...
                    x.start()
                    #self.zmqsender.send_image(self.zmqrpi_name, self.frame)
                    self.zmqlast=zmqnow



        # -------------- timelapse if seconds>0: ******************************* MKV FOR LOOP
        # -------------- timelapse if seconds>0: ******************************* MKV FOR LOOP
        # -------------- timelapse if seconds>0: ******************************* MKV FOR LOOP
        elif seconds>0:
            # ..0.01 means EVERY FRAME ..seconds is in principle different from 'timelapse' value
            WH = (self.frame.shape[1],self.frame.shape[0])

            if not self.aviopened_laps:
                me = socket.gethostname()
                filename = dt.datetime.now().strftime("%Y%m%d_%H%M%S")
                fseconds = int(seconds)
                filename = os.path.expanduser("~/DATA/"+ f"{me}_laps{fseconds:04d}_{filename}.mkv" )
                for i in range(1):
                    print(f"{bg.red}{filename}{bg.default}")

                self.save_avi_counter = 0 # count saved frames

                # https://learn.microsoft.com/en-us/windows/win32/medfound/video-fourccs
                #
                # fourcc = cv2.VideoWriter_fourcc(*'XVID')  # NOT a QUALITY
                # fourcc = cv2.VideoWriter_fourcc(*'IYUV')  # SHOULD BE HUGE....
                fourcc = cv2.VideoWriter_fourcc(*mycodec)  #

                self.aviout_laps = cv2.VideoWriter( filename , fourcc , fps, WH  )
                self.aviopened_laps = True
                # self.avi_last = dt.datetime.now()


            if (dt.datetime.now()-self.avi_last).total_seconds() > seconds:
                # ... second can be 0.001 too ... we pay attention to averaging then
                #      self.accumulate_number = accumulate # info for laps
                #
                # I would like to encode time
                #   DETAILED FONT NO ATIALIAS
                #  ls  -trh | tail -1 | xargs mpv --screenshot-format=png
                #   https://www.dafont.com/small-pixel.font?text=digital
                #
                if seconds<0.1 and  self.accumulate_number>1 and frnum>0:
                    if frnum % self.accumulate_number  != 0:
                        # print(f" NOSAVE    {self.accumulate_number} {frnum}        ")
                        # do not record when accumu
                        return

                self.save_avi_counter+=1

                # # pixel typing
                # for i in range(300,300+ self.save_avi_counter):
                #     self.frame[100,i]=[255,255,255]
                if basecamera_string is not None:
                    # main information to be printed==basecamerastring
                    overtext=basecamera_string #dt.datetime.now().strftime("%H:%M:%S.%f")[:-4] #"1234567890 12:34:55.38   1 2 : 3 4 : 5 5 . 3 8   "
                    overtext = f"{self.save_avi_counter:06d} / {overtext}"
                    fontpath = os.path.expanduser("~/.config/flashcam/small_pixel.ttf")
                    position = ( 480, self.frame.shape[0]-8 ) # 480 on x
                    font = ImageFont.truetype(fontpath, 6)
                    img_pil = Image.fromarray(self.frame)
                    draw = ImageDraw.Draw(img_pil)
                    draw.fontmode = "1" # NO ANTIALIASING
                    overtext = ' '.join(overtext[i:i+1] for i in range(0, len(overtext), 1))
                    drtext =  str(overtext) # to be sure
                    b,g,r,a = 255,255,255,0
                    draw.text( position,  drtext, font = font, fill = (b, g, r, a))
                    self.frame = np.array(img_pil)

                self.aviout_laps.write(self.frame)
                self.avi_last = dt.datetime.now()
                #@{dt.datetime.now().strftime('%H:%M:%S.%f')[:-4]}
                print(f"i... saving timelaps  #{self.save_avi_counter:06d}   [ {basecamera_string} ]" )





#===============================================================
if __name__=="__main__":
    es1 = Stream_Enhancer()
    #if not(es1.add_frame_from_cam()):
    es1.add_frame_from_cam()
    # es1.show_frame()

    print("DD..... RUNALl")
    for i in range(1000):
        for row in range(4): # max 3 rows
            for col in range(6): # 5 columns @ 480 max
                es1.setbox("_Row{}COL{}_".format(row,col), (row,col) )
        for row in range(16): # 5 columns @ 480 max
            es1.setbox("{}COL{}_".format(row,col), (row,col), "left")
        for row in range(16): # 5 columns @ 480 max
            es1.setbox("{}COL{}_".format(row,col), (row,col), "right")

        es1.setbox("MODE avg ",es1.MODE)
        es1.setbox("DISP laps ",es1.DISP)

        es1.setbox("avg = 1 ",es1.avg)
        es1.setbox("blr = 0 ",es1.blr)

        es1.blimp_frame()



# FOR CONNECTION TO FLASHCAM AND NOTIFATOR - SEE FLASHCAM README.org

from console import fg,bg
import importlib
ROOTimported = False
s9009 = None
histogram_v = None
histogram_h = None
try:
    importlib.import_module( "ROOT" )
    ROOTimported = True
    globals()["ROOT"] = importlib.import_module( "ROOT" )
    print(f"i... {fg.green} ROOT     IMPORTED {fg.default}")
except ImportError:
    print(f"i... {fg.red} ROOT NOT IMPORTED {fg.default}")
# finally:
#     print("I... importing",i) # check if fire is imported
#     globals()[i] = importlib.import_module( importnames[i] )
#     print("I... imported:",i)

#import ROOT

import random
import cv2
from flashcam.base_camera2 import BaseCamera

from flashcam.usbcheck import recommend_video
from flashcam.v4lc import  get_resolutions
# import base_camera  #  Switches: slowrate....

import datetime as dt
import time
import socket

import glob

import subprocess as sp
import numpy as np

import flashcam.config as config

from  flashcam.stream_enhancer import Stream_Enhancer


from flashcam import v4lc
from flashcam.v4lc import set_gem, get_gem, tune_histo

from flashcam.mmapwr import mmread_n_clear, mmread

import os
import sys

from notifator import telegram
import threading

from PIL import ImageFont, ImageDraw, Image

from console import fg,bg
import socket

from flashcam.text_write import iprint,fonts_available,\
    get_f_height, get_f_width, get_def_font,\
    dial, signal_strength, text_box

# also in stream_enhancer??!!
SOCKET_GETHOSTNAME =socket.gethostname()

try:
    import pyautogui # take screenshot
except:
    print("X... no DISPLAY, pyautogui cannot be used")
# -----------------------------------------------------------------


current_ip = None

def is_int(n):
    try:
        float_n = float(n)
        int_n = int(float_n)
    except ValueError:
        return False
    else:
        return float_n == int_n

def is_float(n):
    try:
        float_n = float(n)
    except ValueError:
        return False
    else:
        return True

def is_bool(n):
    if type(n) is str and n=="False": return True
    if type(n) is str and n=="True": return True
    return False



# -----------------------------------------------------------------

def get_ip( myip ):
    global current_ip
    if current_ip is None:
        s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        s.connect(("8.8.8.8", 80))
        current_ip = s.getsockname()[0]
        ### current_ip = SOCKET_GETHOSTNAME
        s.close()
    return current_ip

def leftmost_txt_wid( txt ):
    minx = int(640/2)
    lines = str(txt)
    if lines.find("\n")<0:
        lines = [txt]
    else:
        lines = lines.split("\n")
    for i in lines:
        posx = int(640/2 - len( str(i) )*32/2)
        if posx<minx:
            minx = posx
    return minx


def force_text( frame, overtext , posy = 400, bg_bgra = (255,0,255,0) , fg_bgra = (0,255,255,255) , fosize = 2*32):
    """
    7segment LCD text in the middle
    """
    FONT = "utfh"
    midl = 320
    midl = midl - len(str(overtext))*get_f_width(FONT)//2 # center
    frame = iprint( frame, str(overtext), FONT, (midl,posy), fg_bgra)
    return frame
    # posx = leftmost_txt_wid( overtext )# int(640/2 - len( str(overtext) )*32/2)
    # if posx<0: posx=10
    # position = ( posx ,posy)
    # fontpath = os.path.expanduser("~/.config/flashcam/digital-7.mono.ttf")
    # font = ImageFont.truetype(fontpath, fosize)

    # img_pil = Image.fromarray(frame).convert("RGBA")  # base img
    # txt_img = Image.new("RGBA", img_pil.size, bg_bgra)


    # #draw = ImageDraw.Draw(img_pil)
    # draw = ImageDraw.Draw(txt_img)
    # drtext =  f"{overtext}"# {overtextalpha}" # to be sure
    # draw.text( position,  drtext, font = font, fill = fg_bgra )

    # composite = Image.alpha_composite(img_pil, txt_img)
    # #composite.save(output_path)
    # #frame = np.array(img_pil)
    # frame = np.array(composite)
    # return frame

# -----------------------------------------------------------------

class Camera(BaseCamera):

    # video_source = 0
    histomean = 50
    #nfrm = 0 # number frame.... nonono
    # capdevice = None # global

    # CAINI 0
    @staticmethod
    def init_cam(  ):
        """
        should return videocapture device
        but also sould set Camerare.video_source
        """
        global s9009, histogram_v, histogram_h
        #  - all is taken from BaseCam
        # res = "640x480"
        res = config.CONFIG["resolution"]
        print("i... init _ cam caleld with prod:",  config.CONFIG["product"] )

        # here, I need to  wait until the correct video is reported.... ????
        #
        #### HACK ===
        vids = recommend_video( config.CONFIG["product"] , slow_track = False ) # if jpg => give -1
        #
        # if error is lsusb=>>> return [-1] ??????? or [] ???? test it NONONOO
        #

        if len(vids)>0:
            if vids[0]==-1:
                return config.CONFIG["product"] , -1, None

            vidnum = vids[0]
            print("D... asking VideoCapture", vidnum, dt.datetime.now() )
            #
            #  00.0-usb-0:1.3:1.0   00.0-usb-0:1.1:1.0
            #  00.0-usb-0:1.4:1.0   00.0-usb-0:1.2:1.0
            #
            #
            cap = cv2.VideoCapture(vidnum,  cv2.CAP_V4L2)
            #cap = cv2.VideoCapture(vidnum )
            ### cap = cv2.VideoCapture(vidnum,  cv2.CAP_DSHOW) ## ??? ## https://stackoverflow.com/questions/59371075/opencv-error-cant-open-camera-through-video-capture#61817613
            print("D... got    VideoCapture", vidnum , dt.datetime.now())

            # config.CONFIG["camera_on"] = True

            # - with C270 - it showed corrupt jpeg
            # - it allowed to use try: except: and not stuck@!!!
            #cap = cv2.VideoCapture(vidnum)
            #   70% stucks even with timeout


            #pixelformat = "MJPG"

            pixelformat = "YUYV" # I use lossless format for camera readout
            pixelformat = config.CONFIG['PIXELFORMAT']

            time.sleep(0.6)
            fourcc = cv2.VideoWriter_fourcc(*pixelformat) # for capture device
            cap.set(cv2.CAP_PROP_FOURCC, fourcc)
            time.sleep(0.6)

            w,h =  int(res.split("x")[0]), int(res.split("x")[1])
            print(f"i... {fg.green}   RESOLUTIONwh= {w} x {h}, PIXELFORMAT {pixelformat}  {fg.default}")
            cap.set(cv2.CAP_PROP_FRAME_WIDTH,   w )
            time.sleep(0.5)
            cap.set(cv2.CAP_PROP_FRAME_HEIGHT,  h )
            print(f"i... {fg.green}   RESOLUTIONwh= {w} x {h}, PIXELFORMAT {pixelformat}  {fg.default}")

            if ROOTimported and s9009 is None:
                print(f"i... {fg.yellow} ------------------------------ server register {fg.default}")
                histogram_h = ROOT.TH1F("hhtest","hhtest",640,0,640)
                histogram_v = ROOT.TH1F("hvtest","hvtest",480,0,480)
                # serv = ROOT.THttpServer("http:9009?loopback")
                serv = ROOT.THttpServer("http:9009")
                serv.Register("/",histogram_h)
                serv.Register("/",histogram_v)
                serv.SetItemField("/", "_monitoring", "1000")
                serv.SetItemField("/", "_layout", "vert2")
                serv.SetItemField("/", "_drawitem", "[hvtest,hhtest]")


                s9009 = serv
            else:
                serv = None
            return cap,vidnum, serv
        else:
            # returned [] : product not found or a crash in usbls...rpi4b 2 cams
            time.sleep(1)
        return None, None, None

    @staticmethod
    def acquire_one_frame(cap):
        frame = None
        ret = True #  I change it later
        # capture_time=""
        #print("i...acq_onef: cap==", cap)
        if type(cap) == tuple:
            print("i... TUPLE    cap==", cap, "  --->[0]")
            cap = cap[0] # THIS IS STRANGE FOR newcam20211117
            print("i... TUPLE[0] cap==", cap , " type==",type(cap) )
        if (cap is None):
            print("X... cap is none")
            ret = False
        if (cap is not None) and (type(cap) != str) and (not cap.isOpened()):
            print("X... 1camera  not Opened(@real_cam aof)")
            time.sleep(0.7)
        if (cap is not None) and (type(cap) != str) and (not cap.isOpened()):
            print("X... 2camera  not Opened(@real_cam aof)")
            # cap.release() ### NOT TESTED
            ret = False
        if type(cap) == str:
            ret = False
        if ret:
            #print("OK")
            print(f"i... frame {BaseCamera.nframes:8d}  {BaseCamera.capture_time}   ", end="\r" )
            try: #----this catches errors of libjpeg with cv2.CAP_V4L2
                # ret is frequently false on RPI3.... I give 2 more chances
                ret, frame = cap.read()
                if histogram_v is not None and histogram_h is not None:
                    # print(type(frame), len(frame))

                    framered = np.amax(frame, axis=2)

                    suma480 = np.max( framered, axis = 1)
                    suma640 = np.max( framered, axis = 0)

                    #suma480 = np.sum( frame, axis = 1)
                    #suma640 = np.sum( frame, axis = 0)
                    #suma480 =suma480/len(suma480)/3
                    #suma640 =suma640/len(suma640)/3
                    # print(   len(suma480) ,  suma480[100] , np.sum( suma480[100]  ) ) # ax 1 480
                    for i in range(len(suma480) ): # to 480
                        histogram_v.SetBinContent(i, np.sum( suma480[i] ) )
                    for i in range(len(suma640) ): # to 480
                        histogram_h.SetBinContent(i, np.sum( suma640[i] ) )
                        #histogram_v.SetBinContent(i, np.sum( frame, axis = 0)[i] )
                if not ret:
                    print("X... not good read 1 ()")
                    time.sleep(0.1)
                    ret, frame = cap.read()
                    if not ret:
                        print("X... not good read 2")
                        time.sleep(0.2)
                        ret, frame = cap.read()
                        print(f"X... 3rd read result=={ret}")
                        ## cap.release() ## NOT TESTED

                # SIMULATE A PI3B problem
                # frame = cv2.resize(frame, (640,480) )
                if frame is not None:
                    obtres = f"{frame.shape[1]}x{frame.shape[0]}"
                    sucres = obtres == config.CONFIG['resolution']
                    if not sucres:
                        #xzoom
                        #w,h = config.CONFIG['resolution'].split("x")
                        #dsize = ( int(w), int(h))
                        #frame = cv2.resize(frame, dsize )
                        frame[0:4,0:4] = [0,255,255] # set   ;;0 0 255 is RED
                        print(f" OBTAINed res {fg.red}{obtres}{fg.default} is not config res", end = "" )
                    BaseCamera.nframes+=1
                    BaseCamera.capture_time = dt.datetime.now().strftime("%H:%M:%S.%f")[:-4]

            except Exception as ex:
                print("X... SOME EXCEPTION ON cap.read (@real_cam)...", ex)
                config.CONFIG["camera_on"] = False
        # --- camera probably works ret True
        if not ret:
            time.sleep(0.5)
            config.CONFIG["camera_on"] = False
            print("x...  cap didnt go ok, graying... trying to acquire new cap")
            print("x...   here is something i dont get .....why Camera.init ca")
            # static method. should give 3 things
            #    I dont understand this
            #
            # CAINI 1 ?
            cap = Camera.init_cam( ) # WHAT IS THIS? the same?<= static?
            #
            print(f"?... {fg.red}  cap 1 return === {cap} {fg.default}")
            nfrm = 0
            if frame is None:
                height, width = 480, 640
                blank_image = np.zeros((height,width,3), np.uint8)
                # GRAY GREY 255 128 ---- half split image in a case ....happens too frequent..
                blank_image[:,0:width//2] = (90,90,90)      # (B, G, R)
                blank_image[:,width//2:width] = (150,150,150)
                frame = blank_image

            overtext = "XXXXXXXXXXX"
            overtext = get_ip( "" )
            overtext = f"Device lost\non\n{overtext}"
            frame = force_text( frame, overtext, posy = 200)

        return frame, cap



    @staticmethod
    def camera_or_image( cap, vidnum):
        """
        vidnum None or -1 ...   check cap for filenames
        """
        fullpath_fixed_image = "~/.config/flashcam/monoskop.jpg"
        fullpath_fixed_image = os.path.expanduser( fullpath_fixed_image)
        # debug print( cap, vidnum)
        if not os.path.exists(fullpath_fixed_image):
            print("X... monoskop doesnt exist")
            fullpath_fixed_image = None

        if not (vidnum is None) and (vidnum==-1):
            #
            # No videodevice
            #
            # print("i... camera mode : cam-or-image ...  image mode", type(cap), cap, vidnum)
            if (cap.find("screenshot.jpg")==0) and ('pyautogui' in globals()):
                print("i... screenshot mode")
                image = pyautogui.screenshot()
                image = cv2.cvtColor(np.array(image), cv2.COLOR_RGB2BGR)

                scale_percent = 50
                width = int(image.shape[1] * scale_percent / 100)
                height = int(image.shape[0] * scale_percent / 100)
                # dsize
                dsize = (width, height)

                # resize image
                frame = cv2.resize(image, dsize)
                time.sleep(0.3) # from 85%cpu to 20% ??????
                return "image", frame, cap # repeat cap==image

            elif cap.find("clock")==0:
                """
                clock clocks clocks.jpg .....
                """
                #
                # demanded 'living' jpg
                #
                height, width = 480, 640
                blank_image = np.zeros((height,width,3), np.uint8)
                blank_image[0:height//2,0:width//2] = (25,15,15)      # (B, G, R)
                blank_image[0:height//2,width//2:width] = (15,25,15)
                blank_image[height//2:height,width//2:width] = (15,15,25)
                blank_image[height//2:height,0:width//2] = (25,15,25)
                position = (10, 320)
                drtext = dt.datetime.now().strftime("%H:%M:%S.%f")[:-5]
                frame = iprint( blank_image, drtext,
                                "di99" , position,  color_bgra=(0,255,0,200)  )

                time.sleep(0.1) # from 85%cpu to 20% with 0.1;
                return "image", frame, cap # repeat cap==image

            elif cap.find(".jpg")>=0:
                #
                # any JPG
                #
                #print("i... static image mode")
                capfull = os.path.expanduser( f"~/.config/flashcam/{cap}" )
                if  os.path.exists(capfull):
                    fullpath_fixed_image =  capfull
                    #print("W... image        exist", cap, " OK")
                else:
                    print("X... image doesnt exist", cap, "using monoskop")

                time.sleep(0.3) # from 85%cpu to 20% with 0.1
                retimg = cv2.imread( fullpath_fixed_image)
                retimg = cv2.resize(retimg, (640,480) )
                return "image", retimg, cap # repeat cap==image
        else:
            #==================================== monoskop added with IP ADDRESS in black
            # print("i...  camera mode .....   cap-or-image",type(cap), cap, vidnum)
            if cap is None and vidnum is None:
                print("X... camera not accessible ... NO CAMERA FOUND ")
                time.sleep(0.3)
                frame = cv2.resize(cv2.imread( fullpath_fixed_image), (640,480) )
                overtext = "XXXXXXXXXXX"
                overtext = get_ip("")
                overtext = f"No camera found on {overtext}"
                frame = force_text(frame, overtext , posy = 35, fg_bgra = (0,0,0,255) ) # blue
                txt_img = "xxxxxxxxxxx"

                return "image_forced",  frame, cap # repeat cap==image
            # if there is a new cap => propagate it upsrtream
            frame, newcap = Camera().acquire_one_frame(cap)
            return "camera", frame, newcap
        print("X... NEVER GET HERE..................")
        return None, None, None




    @staticmethod
    def frames( ):
        """
        product= ... uses the recommend_video to restart the same cam
        """
        # i need these to be in globals() ----evaluate from web.py
        #                                 ---- OR FROM seread
        global substract_background,save_background
        global save_image_decor, save_image_png # save camera_screenshot - web feature - unlike savebg - it saves with all decorations
        global mix_foreground,save_foreground
        global send_telegram, telegramlast

        global speedx, speedy, restart_translate, average
        global gamma_divide, gamma_multiply,gamma_setdef
        global gain_divide,gain_multiply,gain_setdef
        global expo_divide,expo_multiply,expo_setdef,  expovalue, gainvalue
        global timelaps, rotate180
        global fixed_image # show not camera but image
        global zoom
        global resozoom
        global pausedMOTION
        global overtext, overtextalpha, baduser
        global framekind


        global s9009

        # print("i... staticmethod frames @ real -  enterred; target_frame==", target_frame)
        # ----- I need to inform SENH about the resolution

        senh = Stream_Enhancer( resolution = config.CONFIG["resolution"] ) # i take care inside

        senh.zmqtarget = None # initially
        if 'imagezmq' in config.CONFIG:
            senh.zmqtarget = config.CONFIG['imagezmq']
            if senh.zmqtarget=="None":
                senh.zmqtarget = None
        else:
            print("X... need to update config for imagezmq")
            senh.zmqtarget = None

        if 'jtelegram' in config.CONFIG:
            senh.jtelegram = config.CONFIG['jtelegram']
            if senh.jtelegram=="false":
                senh.jtelegram = False
        else:
            print("X... need to update config for imagezmq")
            senh.jtelegram = False


        # === I must have these GLOBAL and PREDEFINED HERE <= web.py
        # --------------------------------  control
        # -----------get parameters for DetMot, same for web as for all
        #print(config.CONFIG)
        #print( "AVERAGE I AM HAVING ",config.CONFIG['average'] )
        framekind    = config.CONFIG['framekind']
        average      = int(config.CONFIG['average'])
        threshold    = int(config.CONFIG['threshold'])
        blur         = int(config.CONFIG['blur'])
        timelaps     = config.CONFIG['laps']
        histogram    = config.CONFIG['Histogram']
        res          = config.CONFIG['resolution']
        speedx       = float(config.CONFIG['x'])
        speedy       = float(config.CONFIG['y'])
        rotate180    = int(config.CONFIG['otate'])
        zoom         = int(config.CONFIG['zoom'])
        resozoom     = False # trying to play on resolution

        MODE_DMbase = "MODE DM"
        MODE_DM = "MODE DM"

        #imagezmq = None # I use senh.zmqtarget....
        #if 'imagezmq' in config.CONFIG:
        #    imagezmq     = config.CONFIG['imagezmq']

        print( "XY: ", config.CONFIG['x'] ,  config.CONFIG['y']  , speedx, speedy)

        # ------------------    to evaluate commands from web.py
        # ------------------    or searead
        # ------------------    these commands need to be declared here
        #                       AND in globals
        substract_background = False
        save_background = False
        save_image_decor = False # save camera_screenshot
        save_image_png = False # save camera_screenshot
        mix_foreground = False
        save_foreground = False

        send_telegram = None # i hope this is ok too...
        telegramlast = dt.datetime.now()

        restart_translate = False

        gamma_divide = False
        gamma_multiply =False
        gamma_setdef =False

        gain_divide = False
        gain_multiply =False
        gain_setdef =False

        expo_divide = False
        expo_multiply =False
        expo_setdef =False
        exposet = False # not used..
        expovalue = -999. # initial
        gainvalue = -999. # initial

        # rotate180 = False # i define earlier from CONFIG

        fixed_image = None # just camera.

        # --- 433MHz
        pausedMOTION = False
        overtext = None
        overtextalpha = 0
        baduser = None


        switch_res = False # testing
        switch_res_pos = ["C","C"] # initial position of xzoom

        # ==================== GO TO CAMERA AND IMAGE PROCESSING ==============

        camera = Camera(  )
        vidnum = None # it will be re-asked gain and again

        # CAINI 2
        cap, vidnum, s9009 = camera.init_cam(  ) # can return None,None; of jpg,-1
        print("D... CAINI 2 '''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''")
        # Get the current FPS from the camera
        webcam_fps = 10
        try:
            webcam_fps = cap.get(cv2.CAP_PROP_FPS)
        except:
            print("X... no fps available")
        if not webcam_fps or webcam_fps <= 0:
            webcam_fps = 10.0  # Fallback FPS if the camera does not return a valid FPS







        # *********  video works,  get capacities and go with EXPO GAIN
        if not( (vidnum is None) or (vidnum == -1) ):

            cc = v4lc.V4L2_CTL("/dev/video"+str(vidnum))
            capa = cc.get_capbilities()


            #--- INITIATION...... collecting???

            exposuredef = True
            gammadefX = True
            gaindefX = True

            cc.autoexpo_on("autoexpo")
            if "gain" in capa:
                gain = cc.get_gain()
                print(f"========={bg.orange}{fg.black} GAIN INIT  {gain}{fg.default}{bg.default} =","="*40)
                gaindef = cc.getdef_gain()
                if gaindef == gain:
                    gaindefX = True
                else:
                    gaindefX = False

            if "gamma" in capa:
                gamma = cc.get_gamma()
                gammadef = cc.getdef_gamma()
                if gammadef == gamma:
                    gammadefX = True
                else:
                    gammadefX = False

            aea,aex,aga,agm = get_gem(cc, capa)
            if aex!=None: ex,exd,mine,maxe,ex10 = aex
            if agm!=None: gm,gmd,minm,maxm,gm10 = agm
            if aga!=None: ga,gad,ming,maxg,ga10 = aga

            # very stupid camera    ZC0303 Webcam
            if "exposure" in capa:
                exposure = cc.get_exposure()
                exposuredef = cc.getdef_exposure()
                #?????
                #auto_exposuredef = cc.getdef_exposure()
                print(f"i... EXPOAUTO (top) == {exposure} vs def={exposuredef}; ")
                # it seems I lost autoexposure in one RPI4 imx  camera...




            nfrm = 0
            # if config.CONFIG["product"]:
            #     wname = "none "
            # else:
            #     wname = config.CONFIG["product"]
        # ___ exposure and gain stuff here... done _____



        # *********************** INFINITE UNCONDITIONAL LOOP  ****
        # *********************** INFINITE UNCONDITIONAL LOOP  ****
        # *********************** INFINITE UNCONDITIONAL LOOP  ****
        # *********************** INFINITE UNCONDITIONAL LOOP  ****
        frame_prev = None
        while True:

            timeoutok = False
            ret = False
            frame = None

            # can change cap to a new one
            ccoi1, frame, cap = camera.camera_or_image(cap, vidnum)

            #print( frame.shape )
            #print( ccoi1 * 30 )
            ret = True # for the next
            if ccoi1 == "camera" and frame is None:
                ret = False
            elif ccoi1 == "image_forced":
                # CAINI 3
                cap, vidnum, s9009 = camera.init_cam()

                print("D... CAINI 3 '''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''")

                # *********  video works,  get capacities and go with EXPO GAIN
                if not( (vidnum is None) or (vidnum == -1) ):
                    cc = v4lc.V4L2_CTL("/dev/video"+str(vidnum))
                    capa = cc.get_capbilities()

                    exposuredef = True
                    gammadefX = True
                    gaindefX = True

            #print("D... ret==", ret)
            if ret: #********************************************************* operations

                # FIXED IMAGE NOT ALLOWED DURING NORMAL CAMERA MODE - ONLY
                # FIXED IMAGE NOT ALLOWED DURING NORMAL CAMERA MODE - ONLY  from 202304027
                # FIXED IMAGE NOT ALLOWED DURING NORMAL CAMERA MODE - ONLY
                #
                # if fixed_image is not None:
                #     fullpath_fixed_image = "~/.config/flashcam/"+fixed_image
                #     fullpath_fixed_image = os.path.expanduser( fullpath_fixed_image )
                #     if os.path.exists(fullpath_fixed_image):
                #         frame = cv2.imread( fullpath_fixed_image)

                #=====================================================
                #  FIRST OPERATION ON FRAME ...  xzoom  switchresolution
                #  if not 640x480 =>  CUT FROM IT .... when ZoomResolution option is True
                #=====================================================
                if frame.shape[0]!=480 and config.CONFIG['ZoomResolution']:
                    # DO CUT FROM LARGER IMAGE
                    # xzoom thing
                    # print( frame.shape)
                    h,w = frame.shape[:2]  # from 3
                    # get the middle
                    ch = int(h/2)
                    cw = int(w/2)
                    #print(f"D...  xzooming  {switch_res_pos}   " , end="\r")
                    if switch_res_pos[0]=="L": cw=320
                    if switch_res_pos[1]=="U": ch=240

                    if switch_res_pos[0]=="R": cw=w-320
                    if switch_res_pos[1]=="D": ch=h-240

                    # make it 640x480 again :  0:480   0:640    h-480:h  w-640:w
                    frame = frame[ch-240:ch+240, cw-320:cw+320]
                    #else nothing
                frame_prev = frame



                #print(f" (1){fg.cyan}{frame.shape}{fg.default} ", end = "" )
                if senh.add_frame(frame):  # it is a proper image....

                    #=========== BEFORE OTHER === Create final image ====
                    #=========== like ZOOM
                    #=========== THEN CALCULATE HISTO =====
                    #=========== THEN do other stuff


                    # 1. rotate (+translate of the center)
                    # 2. zoom (+translate the center)
                    # 3. histogram !!!here
                    # 4. speed
                    #  others

                    # senh has a frame now
                    if rotate180!=0:   # rotate earlier than zoom
                        senh.rotate180( rotate180 ) #arbitrary int angle

                    if zoom!=1:
                        try:
                            crocfg = os.path.expanduser("~/.config/flashcam/cross.txt")
                            cross_dx, cross_dy  = None, None
                            if os.path.exists(crocfg):
                                with open(crocfg) as f:
                                    cross_dx, cross_dy  = [int(x) for x in next(f).split()]
                                    #senh.zoom( zoom ,0,0 )
                                    senh.zoom( zoom ,cross_dx, cross_dy )
                        except Exception as e:
                            print("!... Problem ar cross.txt file:",e)

                    # ----------  I need to calculate histogram before labels...
                    if histogram: # just calculate a number on plain frame
                        #hmean = senh.histo_mean( ) # hmean STRING NOW
                        hmean = senh.histo_medi( ) # hmean STRING NOW
                        # notwrk #self.histomean = hmean # when called from direct...
                        # print("i... histo value:", hmean)
                        ##tune_histo(cc, hmean )

                    # ---------- before anything - we decode the web command EXECUTE EXECUTION

                    # - compensate for speed of the sky
                    if ((speedx!=0) or (speedy!=0)) \
                    and ((abs(speedx)>1) or (abs(speedy)>1)):
                        senh.translate( speedx, speedy)

                    if restart_translate:
                        senh.reset_camera_start()
                        restart_translate = False


                    # ------------- commands comming from web.py----------------
                    #  expressions     external commands
                    # ------------- COMMANDS COMMING FROM WEB.PY----------------
                    #  expressions
                    # ------------- commands comming from web.py----------------
                    #           -------------- or from seread (fixed_image ...)
                    expression,value = mmread_n_clear( )

                    if expression[:5] != "xxxxx":
                        #print(f"i...  *  EXPR: {expression} == {value}")
                        print(f"i...  *  EXPR: /{expression}/ == /{value}/")
                        print(f"i...  *  EXPR: /{expression}/ == /{value}/")

                        # -------------------- conversions without eval inf float bool, string
                        if is_int(value):
                            print("i... ",value, 'can be safely converted to an integer.')
                            value = int(float(value)) # 1.0 => int crashes
                        elif is_float(value):
                            print("i... ",value, 'is a float with non-zero digit(s) in the fractional-part.')
                            value = float(value)
                        elif is_bool(value):
                            #print("i... ",value, 'is true or false.')
                            if value=="True":
                                value = True
                            else:
                                value = False
                        else:
                            print(f"i... /{value}/ is string, quotes removed.")
                            value = str(value) # i dont care anyway
                            value = value.strip('"').strip("'")

                            #################################################################
                        try: ################################################################
                            # eval makes float float and int int
                            #print("o... evaluating")
                            globals()[expression] = value  #was  eval(value)
                            print("i...                                   expression evaluated:",  globals()[expression])
                            #  the expression MUST BE decleared   in    globals
                        except:
                            print("X... globals expression FAIL",expression,value)

                        # I need a crosscheck here on terminal screen
                        if expression=="overtext":
                            overtextalpha = 0
                        if expression=="baduser":
                            print(f"X... realcam received info: baduser=={value}")
                            baduser = str(value)
                        if (expression=="fixed_image"):
                            if (value is not None) and (value != "None"):
                                print("============== GO TO FIXED IMG ===================",value)
                                fixed_image = value # doesnt worjk anymore
                                config.CONFIG["product"] = value # this may help to switch LIVE to IMG
                                ##camera.init
                                # CAINI 4
                                cap, vidnum, s9009 = camera.init_cam()
                                print("D... CAINI 4 '''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''")

                            else:
                                print("============== RESTOR FROM FIXED IMG ===================")
                                fixed_image = None #value # doesnt worjk anymore
                                config.CONFIG["product"] = "" # this may help to switch LIVE to IMG
                                # CAINI 5
                                cap, vidnum, s9009 = camera.init_cam(  )
                                print("D... CAINI 5 '''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''")



                        if expression=="pausedMOTION":
                            if value is True:
                                print("===================== DM is going STANDBY  ==================************==")
                                if config.CONFIG["cmatrixkill"]:
                                    CMD = "killall cmatrix &"
                                    try:
                                        print("D... killing cmatrix .... there might be no process like that ...")
                                        sp.Popen( CMD, shell=True)
                                    except:
                                        print("X.... cannot kill cmatrix &")
                                else:
                                    print("i... cmatrix is not run nor killed - see the config cfg.json")
                            if value is False:
                                print("--------------------- DM is going to RUN RUN RUN ----------**************--")
                                if config.CONFIG["cmatrixkill"]:
                                    CMD = "xterm -fullscreen -e /usr/bin/cmatrix &"
                                    #CMD = "xterm  -e /usr/bin/cmatrix &"
                                    try:
                                        sp.Popen( CMD, shell=True)
                                    except:
                                        print("X.... cannot xterm  cmatrix")
                                else:
                                    print("i... cmatrix is not run nor killed - see the config cfg.json")



                        if expression=="telegram":
                            print("i... telegram test******************************* value=",value)
                            senh.telegram_send_image(blocking= False) # it has an internal block (300sec)

                        if expression=="timelaps":
                            print("i... TIMELAPS expression******************************* value=",value)
                            #senh.telegram_send_image(blocking= False) # it has an internal block (300sec)




                        # ALREADY EXISTS from webform
                        #if expression == "timelaps":
                        #    print("i... timelaps changes .... ", timelaps. type(timelaps) )
                        #    timelaps = value


                        #
                        # PROBLEM TO SOLVE
                        #    i wait for re-demand from client........ seem to work.....
                        #
                        if expression == "switch_res":
                            #xzoom
                            if type(value)==bool:
                                switch_res = value #not(switch_res)
                                print("i... ok ...ON/OFF")

                                print("i...              EXPERIMENTAL SWITCH RESOLUTION now=", switch_res)
                                if switch_res:
                                    #maxres = "640x480"
                                    maxres = get_resolutions( vidnum )[-1] # from usb_check
                                    config.CONFIG['resolution'] = maxres
                                    print("D.... resolutionswitching to the MAX:", maxres)
                                    try:
                                        width,height = maxres.split("x")
                                        cap.set(cv2.CAP_PROP_FRAME_WIDTH, int(width) )
                                        cap.set(cv2.CAP_PROP_FRAME_HEIGHT, int(height))
                                        print("D.... resolutionswitching to the MAX done:", width,height)
                                    except:
                                        print("X... problem on setting  maxres resolution:", maxres)


                                else:
                                    #xzoom
                                    config.CONFIG['resolution'] = "640x480"
                                    print("D.... resolution to 640x480")
                                    cap.set(cv2.CAP_PROP_FRAME_WIDTH, 640)
                                    cap.set(cv2.CAP_PROP_FRAME_HEIGHT, 480)

                                res = config.CONFIG['resolution']


                            else:
                                #print(f"D...  /{value}/ switch_res  is type",type(switch_res), switch_res, switch_res_pos, switch_res_pos[0], switch_res_pos[1] )
                                if value=="L":
                                    print("val L")
                                    if switch_res_pos[0] == "R":
                                        switch_res_pos[0] = "C"
                                    else:
                                        switch_res_pos[0] = "L"
                                        #print("D... reset sw-re-po:",switch_res_pos[0], switch_res_pos)
                                if value=="R":
                                    if switch_res_pos[0] == "L":
                                        switch_res_pos[0] = "C"
                                    else:
                                        switch_res_pos[0] = "R"

                                if value=="U":
                                    if switch_res_pos[1] == "D":
                                        switch_res_pos[1] = "C"
                                    else:
                                        switch_res_pos[1] = "U"

                                if value=="D":
                                    if switch_res_pos[1] == "U":
                                        switch_res_pos[1] = "C"
                                    else:
                                        switch_res_pos[1] = "D"
                                print("D...  new  switch_res position is ", switch_res_pos )
                                switch_res = True#False # do not imply res.

                            #break # THIS BEAKS INIT, work on PC, not well on RPI


                    if save_image_png:  # camera_screenshot PNG Full quality
                        print("D... HERE I SAVE  image camera_screenshot_PNG Full Quality")
                        if config.CONFIG['datapath'][-1] == "/":
                            pngname = dt.datetime.now().strftime("%Y%m%d_%H%M%S.%f")
                            pngname = pngname[:-5].replace(".","_")
                            cv2.imwrite( config.CONFIG['datapath']+f"camera_screenshot_{pngname}.png" , frame ) # nicely saved
                        else:
                            print("X... You need to specify datapath ending with '/' . No screenshot saved")
                        save_image_png = False

                    if save_background:
                        print("D... HERE I SAVE save_background of mask")
                        print("D... HERE I SAVE save_background of mask")
                        print("D... HERE I SAVE save_background of mask")
                        senh.save_background()
                        save_background = False  # ONE SHOT

                    if substract_background:
                        # print("D... HERE I MUST DO subtraction of mask")
                        # print("D... HERE I MUST DO subtraction of mask")
                        # print("D... HERE I MUST DO subtraction of mask",speedx, speedy)
                        senh.subtract()



                    if save_foreground:
                        print("D... HERE I SAVE save_foreground ")
                        print("D... HERE I SAVE save_foreground ")
                        print("D... HERE I SAVE save_foreground ")
                        senh.save_foreground()
                        save_foreground = False  # ONE SHOT


                    if mix_foreground:
                        # print("D... HERE I mix the foreground")
                        senh.mix()


                    # - compensate for speed of the sky
                    if ((speedx!=0) or (speedy!=0)) and ((abs(speedx)<1) and (abs(speedy)<1)):
                        print(f"speed translate {speedx} {speedy}")
                        senh.translate( speedx, speedy)

                    if restart_translate:
                        senh.reset_camera_start()
                        restart_translate = False

                    # average  THIS IS HERE to be changed TOO (ACCUM)
                    # print("i.... average", average)

                    # timelaps  THIS IS HERE to be changed TOO
                    # print("i.... timelaps", timelaps)

                    #print("i... GAMMAS ", gamma, gammadef )

                    if ccoi1 == "camera":
                        if gamma_divide and "gamma" in capa:
                            gamma_divide = False
                            gammadefX = False
                            gamma = cc.gamma_get("gamma")
                            gamma-=0.1
                            cc.gamma(gamma  )
                            gamma = cc.gamma_get("gamma")
                            v4lc.mk_table(cc)

                        if gamma_multiply and "gamma" in capa:
                            gamma_multiply = False
                            gammadefX = False
                            #if "gamma" in capa:
                            gamma = cc.gamma_get("gamma")
                            gamma+=0.1
                            cc.gamma( gamma )
                            gamma = cc.gamma_get("gamma")
                            v4lc.mk_table(cc)
                            # if gamma!=0:
                            #     newgamma =  int(gamma*2)
                            # else:
                            #     newgamma =  int(1)
                            # cc.set_gamma( newgamma )
                            # gamma = newgamma
                        if gamma_setdef and "gamma" in capa:
                            gamma_setdef = False
                            gammadefX = True
                            gamma = gammadef
                            #if "gamma" in capa:
                            cc.setdef_gamma( )
                            #    gamma = gammadef


                        if gain_divide and "gain" in capa:
                            gain_divide = False
                            gaindefX = False
                            gain = cc.gain_get("gain")
                            gain-=0.1
                            cc.gain(gain)
                            gain = cc.gain_get("gain")
                            v4lc.mk_table(cc)


                        if gain_multiply and "gain" in capa:
                            gain_multiply = False
                            gaindefX = False
                            gain = cc.gain_get("gain")
                            gain+=0.1
                            cc.gain(gain)
                            gain = cc.gain_get("gain")
                            v4lc.mk_table(cc)

                        if gain_setdef and "gain" in capa:
                            gain_setdef = False
                            gaindefX = True
                            gain = gaindef
                            print(f"========={bg.orange}{fg.black} GAIN SETDEF {gain} {fg.default}{bg.default} =","="*40)
                            # if "gain" in capa:
                            cc.setdef_gain( )
                            #   gain = gaindef

                        if "exposure_time_absolute" in capa or "exposure_absolute" in capa:
                            if histogram:
                                if hmean<5:
                                    #print(f"i... BOOSTING EXPOSURE TO 100%")
                                    cc.autoexpo_off( "autoexpo")
                                    cc.expo( 1 ,'expo')     # 0-1 log
                                    print("****** -H param => autoex  ************************ OFF")

                                if hmean>240:
                                    #print(f"i... KILLING MAN EXPOSURE ** TO AUTO,  gain too to avoid problem")
                                    cc.autoexpo_on( "autoexpo")
                                    if "gain" in capa:
                                        cc.setdef_gain()
                                    #exposure = -0.1 + exposure
                                    print("****** -H param => autoex  ********************** ON") # e-a-priority problem

                            if expo_divide:
                                expo_divide = False
                                v4lc.mk_table(cc)
                                exposuredef = False

                                cc.autoexpo_off( "autoexpo")

                                exposure = cc.expo_get("expo_get")
                                #print(f"i ... exposure- = {exposure} ")
                                exposure = -0.1 + exposure
                                cc.expo( exposure ,'expo')     # 0-1 log
                                v4lc.mk_table(cc)

                            if expo_multiply:
                                expo_multiply = False
                                v4lc.mk_table(cc)

                                exposuredef = False

                                # ra = random.uniform(0,1)
                                # print("\n\n", round(ra,3) )
                                cc.autoexpo_off( "autoexpo")
                                exposure = cc.expo_get( 'expo_get')     # 0-1 log
                                #print(" I found exposure === ", exposure)
                                cc.expo( exposure + 0.1 ,'expo')     # 0-1 log
                                v4lc.mk_table(cc)

                            if expo_setdef:
                                expo_setdef = False
                                exposuredef = True
                                v4lc.mk_table(cc)
                                exposure = cc.expo_get("expo_get")
                                print(f"i ... AUTO  was;   exposure = {exposure} ")
                                cc.autoexpo_on()
                                exposure = cc.expo_get("expo_get")
                                print(f"i ... AUTO   is;   exposure = {exposure} ")
                                exposure = 0
                                #senh.setbox(f"expo {exposure:.4f}",  senh.expo)
                                v4lc.mk_table(cc)




                            if not exposuredef: senh.setbox(f"exp {exposure:.3f}",  senh.expo)
                            if not gaindefX: senh.setbox(f"gai {gain:.3f}",  senh.gain)
                            if not gammadefX: senh.setbox(f"gam {gamma:.3f}",  senh.gamma)
                            #if  'exposure' in locals() and exposure != exposuredef:
                            #senh.setbox(f"expo {exposure:.4f}",  senh.expo)
                        #-----------exposure in capa
                    # ______________________ section with capa for camera _____________________


                    #--------------- now apply labels ------i cannot get rid in DETM---
                    #--------- all this will be on all rames histo,detect,direct,delta
                    senh.setbox(" ", senh.TIME, kompr=config.CONFIG['kompress'])
                    if config.CONFIG['resolution'] != "640x480" and config.CONFIG['ZoomResolution']:
                        #xzoom
                        senh.setbox(f"xz{switch_res_pos[0]}{switch_res_pos[1]}", senh.xzoom)

                    #senh.setbox(" ", senh.TIME, kompr= frame.shape[1])

                    if framekind in ["detect","delta","histo"]:
                        senh.setbox(f"DISP {framekind}",senh.DISP)
                    if average>0:
                        senh.setbox(f"a {average}",  senh.avg)
                    if blur>0:
                        senh.setbox(f"b  {blur}",  senh.blr)
                    if threshold>0:
                        senh.setbox(f"t  {threshold}",  senh.trh)
                    if timelaps>0:
                        mycodec=config.CONFIG['FOURCC']
                        if mycodec == "DIVX":  #
                            senh.setbox(f"ld {timelaps}",  senh.lap)
                        elif mycodec == "XDIV":  # not  mkv
                            senh.setbox(f"lx {timelaps}",  senh.lap)
                        elif mycodec == "IYUV":
                            senh.setbox(f"lY {timelaps}",  senh.lap)
                        else:
                            senh.setbox(f"l  {timelaps} {mycodec}",  senh.lap)
                    if timelaps<0:
                        senh.setbox(f"l AS",  senh.lap)

                    if histogram:
                        senh.setbox(f"h {hmean}",  senh.hist)
                    if speedx!=0:
                        #print(speedx)
                        senh.setbox(f"x {speedx:.3f}",  senh.speedx)
                    if speedy!=0:
                        senh.setbox(f"y {speedy:.3f}",  senh.speedy)
                    if zoom!=1:
                        senh.setbox(f"z {zoom:1d}x", senh.scale)

                    if substract_background and not mix_foreground:
                        senh.setbox("-BCKG",  senh.SUBBG )
                    if not substract_background and mix_foreground:
                        senh.setbox("*MIXFG",  senh.SUBBG )
                    if substract_background and mix_foreground:
                        senh.setbox("-BG*FG",  senh.SUBBG )

                    if rotate180!=0:
                        senh.setbox("ROT",  senh.rot )



                    # # ----------------expo gain gamma
                    # # very stupid camera    ZC0303 Webcam
                    # # print(capa, exposure,exposuredef) # crashes
                    # if "exposure" in capa:
                    #     if exposure!=exposuredef: # manual
                    #         senh.setbox(f"expo {exposure}",  senh.expo)

                    # if "auto_exposure" in capa:
                    #     if expo_auto!=expo_autodef: # manual
                    #         senh.setbox(f"expo {exposure_time_absolute}",  senh.expo)

                    # if ("gain" in capa) and (gain!=gaindef): # gain is not frequently tunable
                    #     senh.setbox(f"g {gain}",  senh.gain)

                    # if ("gamma" in capa):
                    #     if (gamma!=gammadef): # manual
                    #         senh.setbox(f"m {gamma}",  senh.gamma)



                    # delayed telegram - preset in DM ========================

                    if not(type(senh.telegramtrigger))==bool:
                        if dt.datetime.now()>senh.telegramtrigger:
                            print("i... telegram time tripped the  2s wire:", senh.telegramtrigger.strftime("%H:%M:%S"), "NOW=",dt.datetime.now().strftime("%H:%M:%S") )
                            senh.telegramtrigger = False
                            senh.telegram_send_image() # it has an internal block (300sec)


                    # ----  for DetMo ---- work with detect motion----------------
                    #   telegram and imagezmq are active only here
                    if (threshold>0) :
                        # here there was MODE DM.
                        # but with imageZMQ and Telegram ALERT....
                        #
                        if not senh.zmqtarget is None:
                            MODE_DM=MODE_DMbase+"z"
                            if senh.jtelegram:
                                MODE_DM=MODE_DM+"T"
                        elif (senh.jtelegram):
                            MODE_DM=MODE_DMbase+"T"


                        senh.setbox(MODE_DM, senh.MODE, grayed = pausedMOTION) #---push UP to avoid DetMot
                        #print("D... detecting motion")
                        senh.detmo( average, blur)
                        senh.chk_threshold( threshold )
                        #
                        # I need a way to block DETMO ....
                        # ??? BLUETOOTH ------- see later
                        #
                        if senh.motion_detected: # saving avi on mation detect
                            # print("D... sav mot", senh.motion_detected)
                            if not pausedMOTION:
                                senh.save_avi( seconds = -1, name = "_dm" , mycodec = config.CONFIG['FOURCC'], container =  config.CONFIG['container'])

                    else:
                        senh.setaccum( average  )
                        senh.setblur( blur )
                        #senh.setbox("MODE  ", senh.MODE)

                    # ---draw histogram
                    #print("                               --- ",framekind)
                    if framekind == "histo":
                        senh.histo( )

                    if timelaps>0:
                        mycodec=config.CONFIG['FOURCC']
                        senh.save_avi( seconds = timelaps,
                                       basecamera_string=f"{BaseCamera.nframes:07d} / {BaseCamera.capture_time}",
                                       mycodec = mycodec, frnum= BaseCamera.nframes)

                    if timelaps<0: # NEW...Save Every frame .. timelaps value is different from seconds value
                        mycodec=config.CONFIG['FOURCC']
                        senh.save_avi( seconds = 0.001,
                                       basecamera_string=f"{BaseCamera.nframes:07d} / {BaseCamera.capture_time}",
                                       mycodec = mycodec, frnum= BaseCamera.nframes)



                    #------------yield the resulting frame-----------------------------
                    if framekind in ["detect","delta","histo"]:
                        frame = senh.get_frame(  typ = framekind)
                    else:
                        frame = senh.get_frame(  )

                    # debug info on final resolution
                    #print(f"  FIN{fg.yellow}{frame.shape}{fg.default} ", end = "" )

                    # --- here I can touch frame:
                    if overtext is not None:
                        frame = force_text( frame, overtext, posy = 400, fg_bgra= (0,255, 0, 255-int(overtextalpha)) )
                        # DIAL --------
                        post = (590,430)
                        frame = dial(frame,post , color=(0,255,0) ,
                                     radius = 30,
                                     thickness = 15,
                                     percentage= 100*overtextalpha/255 )
                        # FO="v14"
                        # frame = iprint( frame,
                        #                 str(round(100*overtextalpha/255)),
                        #                 FO,
                        #                 position = (post[0]-2*get_f_width(FO)//2,
                        #                             post[1] +get_f_height(FO)//2 ),
                        #                 color_bgra= (0,255, 0, 255-int(overtextalpha))
                        #                )

                        frame = signal_strength( frame, position=(580,60) , percentage= 100*overtextalpha/255  )
                        frame = text_box( frame, position=(320,240) )


                        if overtextalpha<255:
                            overtextalpha+=1.2
                        else:
                            overtext = None

                    if baduser is not None:
                        frame = force_text( frame, f"error! => {baduser}", posy = 440, fg_bgra= (0,0,255, 255), fosize = 32 )



                    if save_image_decor:  # camera_screenshot with all decor
                        print("D... HERE I SAVE  image camera_screenshot_decor")
                        if config.CONFIG['datapath'][-1] == "/":
                            cv2.imwrite( config.CONFIG['datapath']+"camera_screenshot.jpg" , frame ) # nicely saved
                        else:
                            print("X... You need to specify datapath ending with '/' . No screenshot saved")
                        save_image_decor = False  # ONE SHOT


            yield frame
